# This file was adapted from scan code definitions in FreeRDP.
# https://github.com/FreeRDP/FreeRDP/blob/master/include/freerdp/scancode.h
#
# FreeRDP: A Remote Desktop Protocol Implementation
# RDP protocol "scancodes"
#
# Copyright 2009-2012 Marc-Andre Moreau <marcandre.moreau@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import namedtuple

ScanCodeTuple = namedtuple("ScanCode", "code extended")

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/089d362b-31eb-4a1a-b6fa-92fe61bb5dbf
KBDFLAGS_EXTENDED = 2


class ScanCode:
    """
    Enumeration for RDP scan codes. Values are a tuple of (scanCode: int, isExtended: bool).
    """

    UNKNOWN = ScanCodeTuple(0x00, False)           # Unknown key
    ESCAPE = ScanCodeTuple(0x01, False)            # VK_ESCAPE
    KEY_1 = ScanCodeTuple(0x02, False)             # VK_KEY_1
    KEY_2 = ScanCodeTuple(0x03, False)             # VK_KEY_2
    KEY_3 = ScanCodeTuple(0x04, False)             # VK_KEY_3
    KEY_4 = ScanCodeTuple(0x05, False)             # VK_KEY_4
    KEY_5 = ScanCodeTuple(0x06, False)             # VK_KEY_5
    KEY_6 = ScanCodeTuple(0x07, False)             # VK_KEY_6
    KEY_7 = ScanCodeTuple(0x08, False)             # VK_KEY_7
    KEY_8 = ScanCodeTuple(0x09, False)             # VK_KEY_8
    KEY_9 = ScanCodeTuple(0x0A, False)             # VK_KEY_9
    KEY_0 = ScanCodeTuple(0x0B, False)             # VK_KEY_0
    OEM_MINUS = ScanCodeTuple(0x0C, False)         # VK_OEM_MINUS
    OEM_PLUS = ScanCodeTuple(0x0D, False)          # VK_OEM_PLUS
    BACKSPACE = ScanCodeTuple(0x0E, False)         # VK_BACK Backspace
    TAB = ScanCodeTuple(0x0F, False)               # VK_TAB
    KEY_Q = ScanCodeTuple(0x10, False)             # VK_KEY_Q
    KEY_W = ScanCodeTuple(0x11, False)             # VK_KEY_W
    KEY_E = ScanCodeTuple(0x12, False)             # VK_KEY_E
    KEY_R = ScanCodeTuple(0x13, False)             # VK_KEY_R
    KEY_T = ScanCodeTuple(0x14, False)             # VK_KEY_T
    KEY_Y = ScanCodeTuple(0x15, False)             # VK_KEY_Y
    KEY_U = ScanCodeTuple(0x16, False)             # VK_KEY_U
    KEY_I = ScanCodeTuple(0x17, False)             # VK_KEY_I
    KEY_O = ScanCodeTuple(0x18, False)             # VK_KEY_O
    KEY_P = ScanCodeTuple(0x19, False)             # VK_KEY_P
    OEM_4 = ScanCodeTuple(0x1A, False)             # VK_OEM_4 '[' on US
    OEM_6 = ScanCodeTuple(0x1B, False)             # VK_OEM_6 ']' on US
    RETURN = ScanCodeTuple(0x1C, False)            # VK_RETURN Normal Enter
    LCONTROL = ScanCodeTuple(0x1D, False)          # VK_LCONTROL
    KEY_A = ScanCodeTuple(0x1E, False)             # VK_KEY_A
    KEY_S = ScanCodeTuple(0x1F, False)             # VK_KEY_S
    KEY_D = ScanCodeTuple(0x20, False)             # VK_KEY_D
    KEY_F = ScanCodeTuple(0x21, False)             # VK_KEY_F
    KEY_G = ScanCodeTuple(0x22, False)             # VK_KEY_G
    KEY_H = ScanCodeTuple(0x23, False)             # VK_KEY_H
    KEY_J = ScanCodeTuple(0x24, False)             # VK_KEY_J
    KEY_K = ScanCodeTuple(0x25, False)             # VK_KEY_K
    KEY_L = ScanCodeTuple(0x26, False)             # VK_KEY_L
    OEM_1 = ScanCodeTuple(0x27, False)             # VK_OEM_1 ';' on US
    OEM_7 = ScanCodeTuple(0x28, False)             # VK_OEM_7 "'" on US
    OEM_3 = ScanCodeTuple(0x29, False)             # VK_OEM_3 Top left, '`' on US, JP DBE_SBCSCHAR
    LSHIFT = ScanCodeTuple(0x2A, False)            # VK_LSHIFT
    OEM_5 = ScanCodeTuple(0x2B, False)             # VK_OEM_5 Next to Enter, '\' on US
    KEY_Z = ScanCodeTuple(0x2C, False)             # VK_KEY_Z
    KEY_X = ScanCodeTuple(0x2D, False)             # VK_KEY_X
    KEY_C = ScanCodeTuple(0x2E, False)             # VK_KEY_C
    KEY_V = ScanCodeTuple(0x2F, False)             # VK_KEY_V
    KEY_B = ScanCodeTuple(0x30, False)             # VK_KEY_B
    KEY_N = ScanCodeTuple(0x31, False)             # VK_KEY_N
    KEY_M = ScanCodeTuple(0x32, False)             # VK_KEY_M
    OEM_COMMA = ScanCodeTuple(0x33, False)         # VK_OEM_COMMA
    OEM_PERIOD = ScanCodeTuple(0x34, False)        # VK_OEM_PERIOD
    OEM_2 = ScanCodeTuple(0x35, False)             # VK_OEM_2 '/' on US
    RSHIFT = ScanCodeTuple(0x36, False)            # VK_RSHIFT
    MULTIPLY = ScanCodeTuple(0x37, False)          # VK_MULTIPLY Numerical
    LMENU = ScanCodeTuple(0x38, False)             # VK_LMENU Left 'Alt' key
    SPACE = ScanCodeTuple(0x39, False)             # VK_SPACE
    CAPSLOCK = ScanCodeTuple(0x3A, False)          # VK_CAPITAL 'Caps Lock', JP DBE_ALPHANUMERIC
    F1 = ScanCodeTuple(0x3B, False)                # VK_F1
    F2 = ScanCodeTuple(0x3C, False)                # VK_F2
    F3 = ScanCodeTuple(0x3D, False)                # VK_F3
    F4 = ScanCodeTuple(0x3E, False)                # VK_F4
    F5 = ScanCodeTuple(0x3F, False)                # VK_F5
    F6 = ScanCodeTuple(0x40, False)                # VK_F6
    F7 = ScanCodeTuple(0x41, False)                # VK_F7
    F8 = ScanCodeTuple(0x42, False)                # VK_F8
    F9 = ScanCodeTuple(0x43, False)                # VK_F9
    F10 = ScanCodeTuple(0x44, False)               # VK_F10
    NUMLOCK = ScanCodeTuple(0x45, False)           # VK_NUMLOCK Note: when this seems to appear in PKBDLLHOOKSTRUCT it means Pause which must be sent as Ctrl + NumLock
    SCROLLLOCK = ScanCodeTuple(0x46, False)        # VK_SCROLL 'Scroll Lock', JP OEM_SCROLL
    NUMPAD7 = ScanCodeTuple(0x47, False)           # VK_NUMPAD7
    NUMPAD8 = ScanCodeTuple(0x48, False)           # VK_NUMPAD8
    NUMPAD9 = ScanCodeTuple(0x49, False)           # VK_NUMPAD9
    SUBTRACT = ScanCodeTuple(0x4A, False)          # VK_SUBTRACT
    NUMPAD4 = ScanCodeTuple(0x4B, False)           # VK_NUMPAD4
    NUMPAD5 = ScanCodeTuple(0x4C, False)           # VK_NUMPAD5
    NUMPAD6 = ScanCodeTuple(0x4D, False)           # VK_NUMPAD6
    ADD = ScanCodeTuple(0x4E, False)               # VK_ADD
    NUMPAD1 = ScanCodeTuple(0x4F, False)           # VK_NUMPAD1
    NUMPAD2 = ScanCodeTuple(0x50, False)           # VK_NUMPAD2
    NUMPAD3 = ScanCodeTuple(0x51, False)           # VK_NUMPAD3
    NUMPAD0 = ScanCodeTuple(0x52, False)           # VK_NUMPAD0
    DECIMAL = ScanCodeTuple(0x53, False)           # VK_DECIMAL Numerical, '.' on US
    SYSREQ = ScanCodeTuple(0x54, False)            # Sys Req
    OEM_102 = ScanCodeTuple(0x56, False)           # VK_OEM_102 Lower left '\' on US
    F11 = ScanCodeTuple(0x57, False)               # VK_F11
    F12 = ScanCodeTuple(0x58, False)               # VK_F12
    SLEEP = ScanCodeTuple(0x5F, False)             # VK_SLEEP OEM_8 on FR (undocumented?)
    ZOOM = ScanCodeTuple(0x62, False)              # VK_ZOOM (undocumented?)
    HELP = ScanCodeTuple(0x63, False)              # VK_HELP (undocumented?)
    F13 = ScanCodeTuple(0x64, False)               # VK_F13 JP agree, should 0x7d according to ms894073
    F14 = ScanCodeTuple(0x65, False)               # VK_F14
    F15 = ScanCodeTuple(0x66, False)               # VK_F15
    F16 = ScanCodeTuple(0x67, False)               # VK_F16
    F17 = ScanCodeTuple(0x68, False)               # VK_F17
    F18 = ScanCodeTuple(0x69, False)               # VK_F18
    F19 = ScanCodeTuple(0x6A, False)               # VK_F19
    F20 = ScanCodeTuple(0x6B, False)               # VK_F20
    F21 = ScanCodeTuple(0x6C, False)               # VK_F21
    F22 = ScanCodeTuple(0x6D, False)               # VK_F22
    F23 = ScanCodeTuple(0x6E, False)               # VK_F23 JP agree
    F24 = ScanCodeTuple(0x6F, False)               # VK_F24 0x87 according to ms894073
    HIRAGANA = ScanCodeTuple(0x70, False)          # JP DBE_HIRAGANA
    HANJA_KANJI = ScanCodeTuple(0x71, False)       # VK_HANJA / VK_KANJI (undocumented?)
    KANA_HANGUL = ScanCodeTuple(0x72, False)       # VK_KANA / VK_HANGUL (undocumented?)
    ABNT_C1 = ScanCodeTuple(0x73, False)           # VK_ABNT_C1 JP OEM_102
    F24_JP = ScanCodeTuple(0x76, False)            # JP F24
    CONVERT_JP = ScanCodeTuple(0x79, False)        # JP VK_CONVERT
    NONCONVERT_JP = ScanCodeTuple(0x7B, False)     # JP VK_NONCONVERT
    TAB_JP = ScanCodeTuple(0x7C, False)            # JP TAB
    BACKSLASH_JP = ScanCodeTuple(0x7D, False)      # JP OEM_5 ('\')
    ABNT_C2 = ScanCodeTuple(0x7E, False)           # VK_ABNT_C2, JP
    HANJA = ScanCodeTuple(0x71, False)             # KR VK_HANJA
    HANGUL = ScanCodeTuple(0x72, False)            # KR VK_HANGUL
    RETURN_KP = ScanCodeTuple(0x1C, True)          # not RETURN Numerical Enter
    RCONTROL = ScanCodeTuple(0x1D, True)           # VK_RCONTROL
    DIVIDE = ScanCodeTuple(0x35, True)             # VK_DIVIDE Numerical
    PRINTSCREEN = ScanCodeTuple(0x37, True)        # VK_EXECUTE/VK_PRINT/VK_SNAPSHOT Print Screen
    RMENU = ScanCodeTuple(0x38, True)              # VK_RMENU Right 'Alt' / 'Alt Gr'
    PAUSE = ScanCodeTuple(0x46, True)              # VK_PAUSE Pause / Break (Slightly special handling)
    HOME = ScanCodeTuple(0x47, True)               # VK_HOME
    UP = ScanCodeTuple(0x48, True)                 # VK_UP
    PRIOR = ScanCodeTuple(0x49, True)              # VK_PRIOR 'Page Up'
    LEFT = ScanCodeTuple(0x4B, True)               # VK_LEFT
    RIGHT = ScanCodeTuple(0x4D, True)              # VK_RIGHT
    END = ScanCodeTuple(0x4F, True)                # VK_END
    DOWN = ScanCodeTuple(0x50, True)               # VK_DOWN
    NEXT = ScanCodeTuple(0x51, True)               # VK_NEXT 'Page Down'
    INSERT = ScanCodeTuple(0x52, True)             # VK_INSERT
    DELETE = ScanCodeTuple(0x53, True)             # VK_DELETE
    NULL = ScanCodeTuple(0x54, True)               # <00>
    HELP2 = ScanCodeTuple(0x56, True)              # Help - documented, different from VK_HELP
    LWIN = ScanCodeTuple(0x5B, True)               # VK_LWIN
    RWIN = ScanCodeTuple(0x5C, True)               # VK_RWIN
    APPS = ScanCodeTuple(0x5D, True)               # VK_APPS Application
    POWER_JP = ScanCodeTuple(0x5E, True)           # JP POWER
    SLEEP_JP = ScanCodeTuple(0x5F, True)           # JP SLEEP
    NUMLOCK_EXTENDED = ScanCodeTuple(0x45, True)   # should be NUMLOCK
    RSHIFT_EXTENDED = ScanCodeTuple(0x36, True)    # should be RSHIFT
    VOLUME_MUTE = ScanCodeTuple(0x20, True)        # VK_VOLUME_MUTE
    VOLUME_DOWN = ScanCodeTuple(0x2E, True)        # VK_VOLUME_DOWN
    VOLUME_UP = ScanCodeTuple(0x30, True)          # VK_VOLUME_UP
    MEDIA_NEXT_TRACK = ScanCodeTuple(0x19, True)   # VK_MEDIA_NEXT_TRACK
    MEDIA_PREV_TRACK = ScanCodeTuple(0x10, True)   # VK_MEDIA_PREV_TRACK
    MEDIA_STOP = ScanCodeTuple(0x24, True)         # VK_MEDIA_MEDIA_STOP
    MEDIA_PLAY_PAUSE = ScanCodeTuple(0x22, True)   # VK_MEDIA_MEDIA_PLAY_PAUSE
    BROWSER_BACK = ScanCodeTuple(0x6A, True)       # VK_BROWSER_BACK
    BROWSER_FORWARD = ScanCodeTuple(0x69, True)    # VK_BROWSER_FORWARD
    BROWSER_REFRESH = ScanCodeTuple(0x67, True)    # VK_BROWSER_REFRESH
    BROWSER_STOP = ScanCodeTuple(0x68, True)       # VK_BROWSER_STOP
    BROWSER_SEARCH = ScanCodeTuple(0x65, True)     # VK_BROWSER_SEARCH
    BROWSER_FAVORITES = ScanCodeTuple(0x66, True)  # VK_BROWSER_FAVORITES
    BROWSER_HOME = ScanCodeTuple(0x32, True)       # VK_BROWSER_HOME
    LAUNCH_MAIL = ScanCodeTuple(0x6C, True)        # VK_LAUNCH_MAIL


"""
Scancode to key name mapping.

Each scancode is an array containing the variant when pressed in
position 0 and the variant when shift is being held in position 1.

# Example
```python
assert SCANCODE_NAMES[0x10][0] == 'q'
assert SCANCODE_NAMES[0x10][1] == 'Q'
```

For scancodes that do not have a different a different name,
`scancode[0] == scancode[1]`
"""
SCANCODE_NAMES = {
    0x01: ['Escape', 'Escape'],
    0x02: ['1',  '!'],
    0x03: ['2',  '@'],
    0x04: ['3',  '#'],
    0x05: ['4',  '$'],
    0x06: ['5',  '%'],
    0x07: ['6',  '^'],
    0x08: ['7',  '&'],
    0x09: ['8',  '*'],
    0x0A: ['9',  '('],
    0x0B: ['0',  ')'],
    0x0C: ['-',  '_'],
    0x0D: ['=',  '+'],
    0x0E: ['Backspace',  'Backspace'],
    0x0F: ['Tab',  'Tab'],
    0x10: ['q',  'Q'],
    0x11: ['w',  'W'],
    0x12: ['e',  'E'],
    0x13: ['r',  'R'],
    0x14: ['t',  'T'],
    0x15: ['y',  'Y'],
    0x16: ['u',  'U'],
    0x17: ['i',  'I'],
    0x18: ['o',  'O'],
    0x19: ['p',  'P'],
    0x1A: ['[',  '{'],
    0x1B: [']',  '}'],
    0x1C: ['Return',  'Return'],
    0x1D: ['Control',  'Control'],
    0x1E: ['a',  'A'],
    0x1F: ['s',  'S'],
    0x20: ['d',  'D'],
    0x21: ['f',  'F'],
    0x22: ['g',  'G'],
    0x23: ['h',  'H'],
    0x24: ['j',  'J'],
    0x25: ['k',  'K'],
    0x26: ['l',  'L'],
    0x27: [';',  ':'],
    0x28: ["'",  '"'],
    0x29: ['`',  '~'],
    0x2A: ['Shift',  'Shift'],
    0x2B: ['\\',  '|'],
    0x2C: ['z',  'Z'],
    0x2D: ['x',  'X'],
    0x2E: ['c',  'C'],
    0x2F: ['v',  'V'],
    0x30: ['b',  'B'],
    0x31: ['n',  'N'],
    0x32: ['m',  'M'],
    0x33: [',',  '<'],
    0x34: ['.',  '>'],
    0x35: ['/',  '?'],
    0x36: ['Shift',  'Shift'],
    0x38: ['Alt',    'AltGr'],
    0x39: ['Space',  'Space'],
    0x3A: ['CapsLock',  'CapsLock'],
    0x3B: ['F1',  'F1'],
    0x3C: ['F2',  'F2'],
    0x3D: ['F3',  'F3'],
    0x3E: ['F4',  'F4'],
    0x3F: ['F5',  'F5'],
    0x40: ['F6',  'F6'],
    0x41: ['F7',  'F7'],
    0x42: ['F8',  'F8'],
    0x43: ['F9',  'F9'],
    0x44: ['F10',  'F10'],
    0x45: ['NumLock',  'NumLock'],
    0x46: ['ScrollLock',  'ScrollLock'],
    0x47: ['Home',  'Home'],
    0x48: ['Up',  'Up'],
    0x49: ['PageUp',  'PageUp'],
    0x4B: ['Left',  'Left'],
    0x4D: ['Right',  'Right'],
    0x4F: ['End',  'End'],
    0x50: ['Down',  'Down'],
    0x51: ['PageDown',  'PageDown'],
    0x52: ['Insert',  'Insert'],
    0x53: ['Delete',  'Delete'],
    0x54: ['SysReq',  'SysReq'],
    0x57: ['F11',  'F11'],
    0x58: ['F12',  'F12'],
    0x5B: ['Meta',  'Meta'],
    0x5D: ['Menu',  'Menu'],
    0x5F: ['Sleep',  'Sleep'],
    0x62: ['Zoom',  'Zoom'],
    0x63: ['Help',  'Help'],
    0x64: ['F13',  'F13'],
    0x65: ['F14',  'F14'],
    0x66: ['F15',  'F15'],
    0x67: ['F16',  'F16'],
    0x68: ['F17',  'F17'],
    0x69: ['F18',  'F18'],
    0x6A: ['F19',  'F19'],
    0x6B: ['F20',  'F20'],
    0x6C: ['F21',  'F21'],
    0x6D: ['F22',  'F22'],
    0x6E: ['F23',  'F23'],
    0x6F: ['F24',  'F24'],
    0x70: ['Hiragana',  'Hiragana'],
    0x71: ['Kanji',  'Kanji'],
    0x72: ['Hangul',  'Hangul'],
    0x5C: ['Windows', 'Windows'],
}


def getKeyName(scanCode: int, isExtended: bool, shiftPressed: bool, capsLockOn: bool) -> str:
    if scanCode in SCANCODE_NAMES:
        code = SCANCODE_NAMES[scanCode]
    else:
        return f"Unknown scan code {hex(scanCode)}"

    key = code[1] if shiftPressed else code[0]
    return f'{key}' if len(key) > 1 else key


def isModifier(scanCode: int):
    """
    Return true if the scancode is one of:
        - Control
        - Alt
        - Shift
        - Windows Key (Meta)
    """
    return scanCode in [0x1D, 0x2A, 0x36, 0x38, 0x5B, 0x5C]



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import logging

# Dependency not installed on Windows. Notifications are not supported
try:
    from pynotifier import Notification
except ImportError:
    pass

class NotifyHandler(logging.StreamHandler):
    """
    Logging handler that sends desktop notifications.
    """

    def __init__(self):
        super(NotifyHandler, self).__init__()

    def emit(self, record):
        """
        Send a notification.
        :param record: the LogRecord object
        """
        try:
            Notification(
                title='PyRDP',
                description=record.getMessage(),
                # duration=5,  # seconds
                urgency='normal'
            ).send()
        except:
            # Either libnotify-bin is not installed or the platform does not support notifications.
            pass


#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
import sys

from progressbar import progressbar

from pyrdp.convert.Converter import Converter
from pyrdp.convert.utils import createHandler
from pyrdp.player import Replay


class ReplayConverter(Converter):
    def process(self):
        with open(self.inputFile, "rb") as f:
            replay = Replay(f)

            print(f"[*] Converting '{self.inputFile}' to {self.format.upper()}")

            outputFileBase = self.outputPrefix + self.inputFile.stem
            handler, outputPath = createHandler(self.format, outputFileBase)

            if not handler:
                print("The input file is already a replay file. Nothing to do.")
                sys.exit(1)

            for event, _ in progressbar(replay):
                handler.onPDUReceived(event)

            print(f"\n[+] Succesfully wrote '{outputPath}'")
            handler.cleanup()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from Crypto.PublicKey.RSA import RsaKey
from Crypto.Util.number import bytes_to_long, long_to_bytes

from pyrdp.security import rc4
from pyrdp.security.key import macData, macSaltedData, generateKeys, updateKey
from pyrdp.enum import EncryptionMethod

"""
Cryptographic utility functions
"""

# Adapted from https://github.com/dlitz/pycrypto/blob/master/lib/Crypto/PublicKey/_slowmath.py
class RSA:
    """
    Class for encrypting or decrypting data with RSA
    Not meant to be a safe implementation. We only need raw RSA without padding.
    """

    def __init__(self, key: RsaKey):
        self.key = key

    def encrypt(self, plaintext: bytes) -> bytes:
        m = bytes_to_long(plaintext)
        ciphertext = pow(m, self.key.e, self.key.n)
        return long_to_bytes(ciphertext)

    def decrypt(self, ciphertext: bytes) -> bytes:
        c = bytes_to_long(ciphertext)

        # compute c**d (mod n)
        if (hasattr(self.key, 'p') and hasattr(self.key, 'q') and hasattr(self.key, 'u')):
            m1 = pow(c, self.key.d % (self.key.p - 1), self.key.p)
            m2 = pow(c, self.key.d % (self.key.q - 1), self.key.q)
            h = m2 - m1

            if (h < 0):
                h = h + self.key.q
            h = h * self.key.u % self.key.q

            plaintext = h * self.key.p + m1
        else:
            plaintext = pow(c, self.key.d, self.key.n)

        return long_to_bytes(plaintext)

class RC4:
    """
    Class for encrypting or decrypting data with RC4.
    """

    def __init__(self, encryptionMethod: EncryptionMethod, macKey: bytes, key: bytes):
        """
        :param encryptionMethod: the encryption method
        :param macKey: the key used for salted signatures.
        :param key: the initial encryption key.
        """
        self.encryptionMethod = encryptionMethod
        self.macKey = macKey
        self.initialBytes = key
        self.currentBytes = key
        self.key = rc4.RC4Key(key)
        self.cipherCount = 0
        self.macCount = 0
    
    def encrypt(self, data: bytes) -> bytes:
        """
        Encrypt data.
        :param data: plaintext data to encrypt.
        :return: encrypted data.
        """
        return rc4.crypt(self.key, data)
    
    def decrypt(self, data: bytes) -> bytes:
        """
        Decrypt data.
        :param data: ciphertext to decrypt.
        :return: decrypted data.
        """
        return self.encrypt(data)
    
    def sign(self, data: bytes, salted: bool) -> bytes:
        """
        Generate a signature for a message.
        :param data: plaintext data to sign.
        :param salted: True if the signature should be salted.
        :return: The signature bytes.
        """
        if salted:
            return macSaltedData(self.macKey, data, self.macCount)[: 8]
        else:
            return macData(self.macKey, data)[: 8]
    
    def verify(self, data: bytes, signature: bytes, salted: bool) -> bool:
        """
        Verify a signature.
        :param data: plaintext data to verify.
        :param signature: the signature to verify.
        :param salted: True if the signature is salted.
        :return: True if the signature is valid.
        """
        return signature[: 8] == self.sign(data, salted)
    
    def increment(self):
        """
        Increment the operation count and update the key if necessary.
        """
        self.cipherCount += 1
        self.macCount += 1

        if self.cipherCount == 4096:
            self.currentBytes = updateKey(self.initialBytes, self.currentBytes, self.encryptionMethod)
            self.key = rc4.RC4Key(self.currentBytes)
            self.cipherCount = 0

class RC4Crypter:
    """
    Class containing RC4 keys for both sides of a communication. Chooses the correct key based on the operation.
    """

    def __init__(self, encryptionMethod: EncryptionMethod, macKey: bytes, encryptKey: bytes, decryptKey: bytes):
        """
        :param encryptionMethod: the encryption method.
        :param macKey: the signing key.
        :param encryptKey: the encryption key.
        :param decryptKey: the decryption key.
        """
        self.encryptionMethod = encryptionMethod
        self.macKey = macKey
        self.encryptKey = RC4(encryptionMethod, macKey, encryptKey)
        self.decryptKey = RC4(encryptionMethod, macKey, decryptKey)

    @staticmethod
    def generateClient(clientRandom: bytes, serverRandom: bytes, encryptionMethod: EncryptionMethod) -> 'RC4Crypter':
        """
        Generate an RC4Crypter instance for RDP clients.
        :param clientRandom: the client random data.
        :param serverRandom: the server random data.
        :param encryptionMethod: the encryption method.
        :return: RC4Crypter
        """
        macKey, decryptKey, encryptKey = generateKeys(clientRandom, serverRandom, encryptionMethod)
        return RC4Crypter(encryptionMethod, macKey, encryptKey, decryptKey)

    @staticmethod
    def generateServer(clientRandom: bytes, serverRandom: bytes, encryptionMethod: EncryptionMethod) -> 'RC4Crypter':
        """
        Generate an RC4Crypter instance for RDP servers.
        :param clientRandom: the client random data.
        :param serverRandom: the server random data.
        :param encryptionMethod: the encryption method.
        :return: RC4Crypter
        """
        macKey, encryptKey, decryptKey = generateKeys(clientRandom, serverRandom, encryptionMethod)
        return RC4Crypter(encryptionMethod, macKey, encryptKey, decryptKey)

    def encrypt(self, data: bytes) -> bytes:
        """
        Encrypt data. The addEncryption method should be called before the next encryption.
        :param data: plaintext to encrypt.
        :return: encrypted data.
        """
        return self.encryptKey.encrypt(data)
    
    def decrypt(self, data: bytes) -> bytes:
        """
        Decrypt data. The addDecryption method should be called before the next decryption.
        :param data: plaintext to decrypt.
        :return: decrypted data.
        """
        return self.decryptKey.decrypt(data)
    
    def sign(self, data: bytes, salted: bool) -> bytes:
        """
        Generate a signature for a message.
        :param data: plaintext to sign.
        :param salted: True if the signature should be salted.
        :return: signature bytes.
        """
        return self.encryptKey.sign(data, salted)
    
    def verify(self, data: bytes, signature: bytes, salted: bool) -> bool:
        """
        Verify a signature for a message.
        :param data: plaintext that was signed.
        :param signature: signature to verify.
        :param salted: True if the signature is salted.
        :return: True if the signature is valid.
        """
        return self.decryptKey.verify(data, signature, salted)
    
    def addEncryption(self):
        """
        Increment the operation count for the encryption key.
        Should be called after each encryption.
        """
        self.encryptKey.increment()
    
    def addDecryption(self):
        """
        Increment the operation count for the decryption key.
        Should be called after each decryption.
        """
        self.decryptKey.increment()

    def getPadLength(self, plaintext: bytes) -> int:
        """
        Get padding length for FIPS. Currently not implemented.
        :param plaintext: plaintext.
        :return: length of padding.
        """
        raise NotImplementedError("FIPS is not implemented")





#!/usr/bin/env python3

#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
# Need to install this reactor before importing any other code
# ruff: noqa: E402
import asyncio
import sys
# We need a special asyncio loop on Windows above Python 3.8. See #316
if (sys.platform == "win32" and sys.version_info.major == 3 and sys.version_info.minor >= 8):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
from twisted.internet import asyncioreactor
asyncioreactor.install(asyncio.new_event_loop())

import logging
import os
import socket

from twisted.internet import reactor

from pyrdp.core.mitm import MITMServerFactory
from pyrdp.mitm.cli import showConfiguration, configure
from pyrdp.logging import LOGGER_NAMES


def main():
    config = configure()
    logger = logging.getLogger(LOGGER_NAMES.MITM_CONNECTIONS)

    # Create a listening socket to accept connections.
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setblocking(0)

    if config.transparent:
        try:
            if not s.getsockopt(socket.SOL_IP, socket.IP_TRANSPARENT):
                s.setsockopt(socket.SOL_IP, socket.IP_TRANSPARENT, 1)
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        except Exception:
            logger.warning('Unable to set transparent socket. Are you running as root?')

    s.bind((config.listenAddress, config.listenPort))
    s.listen()  # Non-blocking.
    reactor.adoptStreamPort(s.fileno(), socket.AF_INET, MITMServerFactory(config))
    s.close()  # reactor creates a copy of the fd.

    message = "MITM Server listening on %(address)s:%(port)d"
    params = {"address": config.listenAddress, "port": config.listenPort}

    if "HOST_IP" in os.environ:
        message += ". Host IP: %(hostIp)s"
        params["hostIp"] = os.environ["HOST_IP"]

    logger.info(message, params)

    reactor.run()

    logger.info("MITM terminated")
    showConfiguration(config)


if __name__ == "__main__":
    main()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

class PyRDPError(Exception):
    """Base class for PyRDP errors"""


class ParsingError(PyRDPError, ValueError):
    def __init__(self, *args):
        super().__init__(*args)
        self.layers = []

    """A parser tried to parse a malformed PDU"""
    def addLayer(self, parser: object, data: bytes):
        self.layers.insert(0, (type(parser).__name__, data))

    def formatLayer(self, index: int) -> str:
        layer = self.layers[index]
        return f"{layer[0]} = {layer[1].hex()}"

    def formatLayers(self) -> str:
        return ",".join(self.formatLayer(i) for i in range(len(self.layers)))


class ExploitError(ParsingError):
    """
    Class used when an exploit attempt or scan is detected in a parser and we want to shut down the connection.
    """

    def __init__(self, *args):
        super().__init__(*args)


class WritingError(PyRDPError, ValueError):
    """A parser tried to write a malformed PDU"""


class UnknownPDUTypeError(PyRDPError, NotImplementedError):
    """A parser tried to write or parse an unknown PDU type"""

    def __init__(self, message, type):
        super(UnknownPDUTypeError, self).__init__(message)
        self.type = type


class StateError(PyRDPError, RuntimeError):
    """Used when trying to do something that an object's state does not allow"""


class CrypterUnavailableError(PyRDPError):
    """Used when trying to use a CrypterProxy before the actual crypter was generated"""



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import asyncio
from typing import Callable, Union
import socket
import logging

from twisted.internet.protocol import ClientFactory, Protocol
from twisted.internet import tcp, fdesc, reactor

LOG = logging.getLogger(__name__)


class AwaitableClientFactory(ClientFactory):
    """
    Twisted Client Factory with an asyncio.Event that is set when the connection is established.
    """

    def __init__(self, protocol: Union[Protocol, Callable[[], Protocol]]):
        """
        :param protocol: protocol to return in buildProtocol, or callable that returns a Protocol object.
        """

        self.protocol = protocol
        self.connected = asyncio.Event()

    def buildProtocol(self, addr):
        self.connected.set()

        if callable(self.protocol):
            return self.protocol()
        else:
            return self.protocol


class TransparentClient(tcp.Client):
    """A TCP client that supports transparent proxying."""

    def createInternetSocket(self):
        """Overridden"""

        err = None
        s = socket.socket(self.addressFamily, self.socketType)
        s.setblocking(0)
        fdesc._setCloseOnExec(s.fileno())

        try:
            if not s.getsockopt(socket.SOL_IP, socket.IP_TRANSPARENT):
                s.setsockopt(socket.SOL_IP, socket.IP_TRANSPARENT, 1)
        except Exception as e:
            LOG.error('Failed to establish transparent proxy: %s', e)

        return s  # Maintain non-transparent behavior.


class TransparentConnector(tcp.Connector):
    """A TCP connector which creates TransparentClients."""

    def _makeTransport(self):
        return TransparentClient(self.host, self.port, self.bindAddress, self, self.reactor)


def connectTransparent(host, port, factory, timeout=30, bindAddress=None):
    """Create a transparent proxy connection managed by Twisted."""
    c = TransparentConnector(host, port, factory, timeout, bindAddress, reactor)
    c.connect()
    return c



#!/usr/bin/env python3

#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
# Need to install this reactor before importing any other code
# ruff: noqa: E402
import asyncio
import sys
# We need a special asyncio loop on Windows above Python 3.8. See #316
if (sys.platform == "win32" and sys.version_info.major == 3 and sys.version_info.minor >= 8):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
from twisted.internet import asyncioreactor
asyncioreactor.install(asyncio.new_event_loop())

from pathlib import Path
import argparse
import logging
import logging.handlers
import os
import sys

from pyrdp.core import settings  # noqa
from pyrdp.logging import LOGGER_NAMES, NotifyHandler, configure as configureLoggers  # noqa
from pyrdp.player import HAS_GUI  # noqa
from pyrdp.player.config import DEFAULTS  # noqa


# Workaround a macOS bug: https://bugreports.qt.io/browse/QTBUG-87014
os.environ['QT_MAC_WANTS_LAYER']='1'
if HAS_GUI:
    from pyrdp.player import MainWindow
    from PySide6.QtWidgets import QApplication


def enableNotifications(logger):
    """Enable notifications if supported."""
    # https://docs.python.org/3/library/os.html
    if os.name != "nt" and sys.platform != "darwin":
        notifyHandler = NotifyHandler()
        notifyHandler.setFormatter(logging.Formatter("[{asctime}] - {message}", style="{"))

        uiLogger = logging.getLogger(LOGGER_NAMES.PLAYER_UI)
        uiLogger.addHandler(notifyHandler)
    else:
        logger.warning("Notifications are not supported on this platform.")


def main():
    """
    Parse the provided command line arguments and launch the GUI.
    :return: The app exit code (0 for normal exit, non-zero for errors)
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("replay", help="Replay files to open on launch (optional)", nargs="*")
    parser.add_argument("-b", "--bind", help="Bind address (default: 127.0.0.1)", default="127.0.0.1")
    parser.add_argument("-p", "--port", help="Bind port (default: 3000)", default=3000)
    parser.add_argument("-o", "--output", help="Output folder", default="pyrdp_output")
    parser.add_argument("-L", "--log-level", help="Log level", default=None,
                        choices=["INFO", "DEBUG", "WARNING", "ERROR", "CRITICAL"], nargs="?")
    parser.add_argument("-F", "--log-filter",
                        help="Only show logs from this logger name (accepts '*' wildcards)", default=None)
    parser.add_argument("--headless", help="Parse a replay without rendering the user interface.", action="store_true")
    args = parser.parse_args()

    cfg = settings.load(f'{settings.CONFIG_DIR}/player.ini', DEFAULTS)

    # Modify configuration with switches.
    if args.log_level:
        cfg.set('vars', 'level', args.log_level)
    if args.log_filter:
        cfg.set('logs', 'filter', args.log_filter)
    if args.output:
        cfg.set('vars', 'output_dir', args.output)

    outDir = Path(cfg.get('vars', 'output_dir'))
    outDir.mkdir(exist_ok=True)

    configureLoggers(cfg)
    logger = logging.getLogger(LOGGER_NAMES.PYRDP)

    if cfg.getboolean('logs', 'notifications', fallback=False) and not args.headless:
        enableNotifications(logger)

    if not HAS_GUI and not args.headless:
        logger.error('Headless mode is not specified and PySide6 is not installed.'
                     ' Install PySide6 to use the graphical user interface.')
        sys.exit(127)

    if not args.headless:
        app = QApplication(sys.argv)
        mainWindow = MainWindow(args.bind, int(args.port), args.replay)
        mainWindow.showMaximized()
        mainWindow.show()

        return app.exec_()
    else:
        logger.info('Starting PyRDP Player in headless mode.')
        from pyrdp.player import HeadlessEventHandler
        from pyrdp.player.Replay import Replay
        eventHandler = HeadlessEventHandler()

        for replayPath in args.replay:
            eventHandler.output.write(f'== REPLAY FILE: {replayPath}\n')

            with open(replayPath, "rb") as f:
                replay = Replay(f)

                for event, _ in replay:
                    eventHandler.onPDUReceived(event)

            eventHandler.output.write('\n-- END --------------------------------\n')


if __name__ == '__main__':
    sys.exit(main())



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum, Flag


class NegotiationType(IntEnum):
    """
    Negotiation data structure type.
    """
    TYPE_RDP_NEG_REQ = 0x01
    TYPE_RDP_NEG_RSP = 0x02
    TYPE_RDP_NEG_FAILURE = 0x03
    TYPE_RDP_CORRELATION_INFO = 0x06


class NegotiationRequestFlags(IntEnum):
    """
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/902b090b-9cb3-4efc-92bf-ee13373371e3
    """
    NONE = 0x00
    RESTRICTED_ADMIN_MODE_REQUIRED = 0x01
    REDIRECTED_AUTHENTICATION_MODE_REQUIRED = 0x02
    CORRELATION_INFO_PRESENT = 0x08


#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Optional

from pyrdp.enum import NTLMSSPMessageType
from pyrdp.pdu import NTLMSSPAuthenticatePDU, NTLMSSPChallengePDU, NTLMSSPNegotiatePDU, NTLMSSPPDU


class NTLMSSPState:
    def __init__(self):
        self.negotiate: Optional[NTLMSSPNegotiatePDU] = None
        self.challenge: Optional[NTLMSSPChallengePDU] = None
        self.authenticate: Optional[NTLMSSPAuthenticatePDU] = None

    def setMessage(self, pdu: NTLMSSPPDU):
        if pdu.messageType == NTLMSSPMessageType.NEGOTIATE_MESSAGE:
            self.negotiate = pdu
        elif pdu.messageType == NTLMSSPMessageType.CHALLENGE_MESSAGE:
            self.challenge = pdu
        else:
            self.authenticate = pdu




#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Dict

from pyrdp.core import ObservedBy, Uint8
from pyrdp.enum import SegmentationPDUType
from pyrdp.layer.buffered import BufferedLayer
from pyrdp.layer.layer import BaseLayer, LayerObserver


class SegmentationObserver(LayerObserver):
    def onUnknownHeader(self, header, data: bytes):
        pass


@ObservedBy(SegmentationObserver)
class SegmentationLayer(BaseLayer):
    """
    Layer to handle segmentation PDUs (e.g: TPKT and fast-path).
    Sends data to the proper BufferedLayer by checking the PDU's header.
    """

    def __init__(self):
        super().__init__()
        self.layers: Dict[int, BufferedLayer] = {}

    def attachLayer(self, pduType: int, layer: BufferedLayer):
        """
        Set the layer used for a type of segmentation PDU.
        :param pduType: the PDU type.
        :param layer: the layer to use.
        """
        # The segmentation layer is bypassed when sending data.
        layer.previous = self.previous
        self.layers[pduType] = layer

    def recv(self, data: bytes):
        """
        Forward data to the proper layer depending on the PDU type.qq
        """

        while len(data) > 0:
            layer = None
            length = 0

            # Check if any layer still needs to receive more data.
            for _, layer in self.layers.items():
                length = layer.getDataLengthRequired()

                if length > 0:
                    break

            if length == 0:
                # All layers are clear, this is a new PDU.
                # The PDU type is contained within the first byte of every message.
                header = Uint8.unpack(data[0]) & SegmentationPDUType.MASK

                try:
                    layer = self.layers[header]
                except KeyError:
                    if self.observer:
                        self.observer.onUnknownHeader(header, data)
                        return
                    else:
                        raise

                layer.recv(data[0 : 1])
                data = data[1 :]
                length = layer.getDataLengthRequired()

            # Send data to the selected layer as long as it needs some and as long as we have data to send.
            while length > 0 and len(data) > 0:
                forwarded = data[: length]
                data = data[length :]
                layer.recv(forwarded)
                length = layer.getDataLengthRequired()


#
# Copyright (c) 2014-2015 Sylvain Peyrefitte
#
# This file is part of rdpy.
#
# rdpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import hashlib

from pyrdp.core import ByteStream, Uint32LE
from pyrdp.enum import EncryptionMethod
from pyrdp.security import rc4


def saltedHash(inputData, salt, salt1, salt2):
    """
    Generate a signature from a combination of sha1 and md5.
    Signature = MD5(Salt + SHA1(Input + Salt + Salt1 + Salt2))
    See: http://msdn.microsoft.com/en-us/library/cc241992.aspx
    :param inputData: strange input (see doc)
    :type inputData: bytes
    :param salt: salt for context call
    :type salt: bytes
    :param salt1: another salt (ex : client random)
    :type salt1: bytes
    :param salt2: another salt (ex: server random)
    :type salt2: bytes
    :return: str
    """
    sha1Digest = hashlib.sha1()
    md5Digest = hashlib.md5()

    sha1Digest.update(inputData)
    sha1Digest.update(salt[:48])
    sha1Digest.update(salt1)
    sha1Digest.update(salt2)
    sha1Sig = sha1Digest.digest()

    md5Digest.update(salt[:48])
    md5Digest.update(sha1Sig)

    return md5Digest.digest()


def finalHash(key, random1, random2):
    """
    Hash = MD5(in0[:16] + in1[:32] + in2[:32])
    :param key: 16 byte string
    :type key: bytes
    :param random1: 32 byte random string
    :type random1: bytes
    :param random2: 32 byte random string
    :type random2: bytes
    :return: MD5(in0[:16] + in1[:32] + in2[:32])
    """
    md5Digest = hashlib.md5()
    md5Digest.update(key)
    md5Digest.update(random1)
    md5Digest.update(random2)
    return md5Digest.digest()


def generateMasterSecret(preMasterSecret, clientRandom, serverRandom):
    """
    Generate master secret.
    See: http://msdn.microsoft.com/en-us/library/cc241992.aspx
    :param preMasterSecret: secret
    :type preMasterSecret: bytes
    :param clientRandom: client random
    :type clientRandom: bytes
    :param serverRandom: server random
    :type serverRandom: bytes
    """
    return saltedHash(b"A", preMasterSecret, clientRandom, serverRandom) + saltedHash(b"BB", preMasterSecret, clientRandom, serverRandom) + saltedHash(b"CCC", preMasterSecret, clientRandom, serverRandom)


def generateSessionKeyBlob(masterSecret, clientRandom, serverRandom):
    """
    Generate session key blob.
    :param masterSecret: secret
    :type masterSecret: bytes
    :param clientRandom: client random
    :type clientRandom: bytes
    :param serverRandom: server random
    :type serverRandom: bytes
    """
    return saltedHash(b"X", masterSecret, clientRandom, serverRandom) + saltedHash(b"YY", masterSecret, clientRandom, serverRandom) + saltedHash(b"ZZZ", masterSecret, clientRandom, serverRandom)


def macData(macKey, data):
    """
    Generate an unsalted signature.
    See: http://msdn.microsoft.com/en-us/library/cc241995.aspx
    :param macKey: signing key.
    :type macKey: bytes
    :param data: data to sign.
    :type data: bytes
    :return: str
    """
    sha1Digest = hashlib.sha1()
    md5Digest = hashlib.md5()

    dataLength = ByteStream()
    Uint32LE.pack(len(data), dataLength)

    sha1Digest.update(macKey)
    sha1Digest.update(b"\x36" * 40)
    sha1Digest.update(dataLength.getvalue())
    sha1Digest.update(data)

    sha1Sig = sha1Digest.digest()

    md5Digest.update(macKey)
    md5Digest.update(b"\x5c" * 48)
    md5Digest.update(sha1Sig)

    return md5Digest.digest()


def macSaltedData(macKey, data, encryptionCount):
    """
    Generate a salted signature.
    See: https://msdn.microsoft.com/en-us/library/cc240789.aspx
    :param macKey: signing key.
    :type macKey: bytes
    :param data: data to sign.
    :type data: bytes
    :param encryptionCount: the number of encrypted packets so far.
    :type encryptionCount: int
    :return: str
    """
    sha1Digest = hashlib.sha1()
    md5Digest = hashlib.md5()

    dataLengthS = ByteStream()
    Uint32LE.pack(len(data), dataLengthS)

    encryptionCountS = ByteStream()
    Uint32LE.pack(encryptionCount, encryptionCountS)

    sha1Digest.update(macKey)
    sha1Digest.update(b"\x36" * 40)
    sha1Digest.update(dataLengthS.getvalue())
    sha1Digest.update(data)
    sha1Digest.update(encryptionCountS.getvalue())

    sha1Sig = sha1Digest.digest()

    md5Digest.update(macKey)
    md5Digest.update(b"\x5c" * 48)
    md5Digest.update(sha1Sig)

    return md5Digest.digest()


def tempKey(initialKey, currentKey):
    """
    Create a temporary updated key.
    See: http://msdn.microsoft.com/en-us/library/cc240792.aspx
    :param initialKey: starting key.
    :param currentKey: current key.
    :return: str
    """
    sha1Digest = hashlib.sha1()
    md5Digest = hashlib.md5()

    sha1Digest.update(initialKey)
    sha1Digest.update(b"\x36" * 40)
    sha1Digest.update(currentKey)

    sha1Sig = sha1Digest.digest()

    md5Digest.update(initialKey)
    md5Digest.update(b"\x5c" * 48)
    md5Digest.update(sha1Sig)

    return md5Digest.digest()


def gen40bits(data):
    """
    Generate 40 bits of data from 128 bits of data for key computation.
    See: http://msdn.microsoft.com/en-us/library/cc240785.aspx
    :param data: 128 bits string
    :type data: bytes
    :return: dict
    """
    return b"\xd1\x26\x9e" + data[:8][-5:]


def gen56bits(data):
    """
    Generate 56 bits of data from 128 bits of data for key computation.
    See: http://msdn.microsoft.com/en-us/library/cc240785.aspx
    :param data: 128 bits string
    :type data: bytes
    :return: str
    """
    return b"\xd1" + data[:8][-7:]


def generateKeys(clientRandom, serverRandom, method):
    """
    Generate cryptographic keys based on client and server random values and a method.
    Returns the mac key, the server-client encryption key and the client-server encryption key.
    See: http://msdn.microsoft.com/en-us/library/cc240785.aspx
    :param method: the EncryptionMethod value.
    :type method: EncryptionMethod
    :param clientRandom: the client random.
    :type clientRandom: bytes
    :param serverRandom: the server random.
    :type serverRandom: bytes
    :return: str, str, str
    """
    preMasterHash = clientRandom[:24] + serverRandom[:24]
    masterHash = generateMasterSecret(preMasterHash, clientRandom, serverRandom)
    sessionKey = generateSessionKeyBlob(masterHash, clientRandom, serverRandom)
    macKey128 = sessionKey[:16]
    initialFirstKey128 = finalHash(sessionKey[16: 32], clientRandom, serverRandom)
    initialSecondKey128 = finalHash(sessionKey[32: 48], clientRandom, serverRandom)

    if method == EncryptionMethod.ENCRYPTION_40BIT:
        return gen40bits(macKey128), gen40bits(initialFirstKey128), gen40bits(initialSecondKey128)
    elif method == EncryptionMethod.ENCRYPTION_56BIT:
        return gen56bits(macKey128), gen56bits(initialFirstKey128), gen56bits(initialSecondKey128)
    elif method == EncryptionMethod.ENCRYPTION_128BIT:
        return macKey128, initialFirstKey128, initialSecondKey128

    raise ValueError("Bad encryption method")


def updateKey(initialKey, currentKey, method):
    """
    Update a key.
    See: http://msdn.microsoft.com/en-us/library/cc240792.aspx
    :param initialKey: initial key.
    :type initialKey: bytes
    :param currentKey: current key.
    :type currentKey: bytes
    :param method: encryption method.
    :type method: EncryptionMethod
    :return: str
    """
    if method == EncryptionMethod.ENCRYPTION_40BIT:
        tempKey128 = tempKey(initialKey[:8], currentKey[:8])
        return gen40bits(rc4.crypt(rc4.RC4Key(tempKey128[:8]), tempKey128[:8]))
    elif method == EncryptionMethod.ENCRYPTION_56BIT:
        tempKey128 = tempKey(initialKey[:8], currentKey[:8])
        return gen56bits(rc4.crypt(rc4.RC4Key(tempKey128[:8]), tempKey128[:8]))
    elif method == EncryptionMethod.ENCRYPTION_128BIT:
        tempKey128 = tempKey(initialKey, currentKey)
        return rc4.crypt(rc4.RC4Key(tempKey128), tempKey128)


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum


class GCCPDUType(IntEnum):
    """
    PDU types for GCC messages received in MCS Connect Initial PDUs.
    """
    CREATE_CONFERENCE_REQUEST = 0
    CREATE_CONFERENCE_RESPONSE = 0x14


#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from PySide6.QtCore import QObject, Qt
from PySide6.QtGui import QCloseEvent
from PySide6.QtWidgets import QDialog, QLabel, QMessageBox, QProgressBar, QPushButton, QVBoxLayout


class FileDownloadDialog(QDialog):
    def __init__(self, remotePath: str, targetPath: str, isMultipleDownload: bool, parent: QObject):
        super().__init__(parent, Qt.CustomizeWindowHint | Qt.WindowTitleHint)
        self.titleLabel = QLabel(f"Downloading {remotePath} to {targetPath}")

        self.progressBar = QProgressBar()
        self.progressBar.setMinimum(0)
        self.progressBar.setMaximum(0)


        self.actualProgress = 0
        self.actualMaximum = 0
        self.isComplete = False

        self.isMultipleDownload = isMultipleDownload
        self.downloadCount = 0
        self.downloadTotal = 0

        self.progressLabel = QLabel(f"{self.downloadCount} / {self.downloadTotal} files downloaded")
        self.progressSizeLabel = QLabel("0 bytes")

        self.widgetLayout = QVBoxLayout()
        self.widgetLayout.addWidget(self.titleLabel)
        self.widgetLayout.addWidget(self.progressLabel)
        self.widgetLayout.addWidget(self.progressBar)
        self.widgetLayout.addWidget(self.progressSizeLabel)

        self.closeButton = QPushButton("Continue download in background")
        self.closeButton.clicked.connect(self.hide)
        self.widgetLayout.addWidget(self.closeButton)

        self.setLayout(self.widgetLayout)

    def getHumanReadableSize(self, size: int):
        prefixes = ["", "K", "M", "G"]

        while len(prefixes) > 1:
            if size < 1024:
                break

            prefixes.pop(0)
            size /= 1024

        return f"{'%.2f' % size if size % 1 != 0 else int(size)} {prefixes[0]}"

    def incrementDownloadCount(self):
        self.downloadCount += 1
        self.progressLabel.setText(f"{self.downloadCount} / {self.downloadTotal} files downloaded")

    def incrementDownloadTotal(self):
        self.downloadTotal += 1
        self.progressLabel.setText(f"{self.downloadCount} / {self.downloadTotal} files downloaded")

    def updateProgress(self):
        progress = self.getHumanReadableSize(self.actualProgress)

        if self.actualMaximum > 0:
            percentage = int(self.actualProgress / self.actualMaximum * 100)
            maximum = self.getHumanReadableSize(self.actualMaximum)

            self.progressBar.setValue(percentage)
            self.progressSizeLabel.setText(f"{progress}B / {maximum}B ({percentage}%)")
        else:
            self.progressBar.setValue(0)
            self.progressSizeLabel.setText(f"{progress}B")

    def reportSize(self, maximum: int):
        self.actualMaximum = maximum

        if self.actualMaximum == 0:
            self.progressBar.setMaximum(0)
        else:
            self.progressBar.setMaximum(100)

        self.updateProgress()

    def reportProgress(self, progress: int):
        self.actualProgress += progress
        self.updateProgress()

    def reportCompletion(self, error: int):
        self.show()

        # QMessageBox is a modal, meaning that execution of code stops until the window closes
        # We need our own message box if we want to queue multiple downloads
        box = QMessageBox(self)
        box.setWindowTitle("Download Complete")
        box.setWindowModality(Qt.NonModal)

        if error == 0:
            box.setIcon(QMessageBox.Information)
            box.setText("Download completed successfully.")
        else:
            box.setIcon(QMessageBox.Critical)
            box.setText(f"Download failed. Error code: {'0x%08lx' % error}")

        box.show()
        self.isComplete = True
        self.close()

    def closeEvent(self, event: QCloseEvent):
        if not self.isComplete:
            event.ignore()



#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
Enumerations for Drawing Orders.
"""

from enum import IntEnum


class DrawingOrderControlFlags(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241574.aspx
    """
    TS_STANDARD = 0x01
    TS_SECONDARY = 0x02
    TS_BOUNDS = 0x04
    TS_TYPE_CHANGE = 0x08
    TS_DELTA_COORDS = 0x10
    TS_ZERO_BOUNDS_DELTAS = 0x20
    TS_ZERO_FIELD_BYTE_BIT0 = 0x40
    TS_ZERO_FIELD_BYTE_BIT1 = 0x80


class Primary(IntEnum):
    DSTBLT = 0x00
    PATBLT = 0x01
    SCRBLT = 0x02
    DRAW_NINE_GRID = 0x07
    MULTI_DRAW_NINE_GRID = 0x08
    LINE_TO = 0x09
    OPAQUE_RECT = 0x0A
    SAVE_BITMAP = 0x0B
    MEMBLT = 0x0D
    MEM3BLT = 0x0E
    MULTI_DSTBLT = 0x0F
    MULTI_PATBLT = 0x10
    MULTI_SCRBLT = 0x11
    MULTI_OPAQUE_RECT = 0x12
    FAST_INDEX = 0x13
    POLYGON_SC = 0x14
    POLYGON_CB = 0x15
    POLYLINE = 0x16
    FAST_GLYPH = 0x18
    ELLIPSE_SC = 0x19
    ELLIPSE_CB = 0x1A
    GLYPH_INDEX = 0x1B


class Secondary(IntEnum):
    BITMAP_UNCOMPRESSED = 0x00
    CACHE_COLOR_TABLE = 0x01
    CACHE_BITMAP_COMPRESSED = 0x02
    CACHE_GLYPH = 0x03
    BITMAP_UNCOMPRESSED_V2 = 0x04
    BITMAP_COMPRESSED_V2 = 0x05
    CACHE_BRUSH = 0x07
    BITMAP_COMPRESSED_V3 = 0x08


class Alternate(IntEnum):
    SWITCH_SURFACE = 0x00
    CREATE_OFFSCREEN_BITMAP = 0x01
    STREAM_BITMAP_FIRST = 0x02
    STREAM_BITMAP_NEXT = 0x03
    CREATE_NINE_GRID_BITMAP = 0x04
    GDIPLUS_FIRST = 0x05
    GDIPLUS_NEXT = 0x06
    GDIPLUS_END = 0x07
    GDIPLUS_CACHE_FIRST = 0x08
    GDIPLUS_CACHE_NEXT = 0x09
    GDIPLUS_CACHE_END = 0x0A
    WINDOW = 0x0B
    COMPDESK_FIRST = 0x0C
    FRAME_MARKER = 0x0D



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import asyncio
from queue import Queue
from typing import Dict

from PySide6.QtCore import Qt, Signal
from PySide6.QtWidgets import QApplication, QFileIconProvider, QMessageBox, QWidget

from pyrdp.player.BaseWindow import BaseWindow
from pyrdp.player.LiveTab import LiveTab
from pyrdp.player.LiveThread import LiveThread


class LiveWindow(BaseWindow):
    """
    Class that holds logic for live player (network RDP connections as they happen) tabs.
    """

    connectionReceived = Signal()
    closedTabText = " - Closed"

    def __init__(self, address: str, port: int, updateCountSignal: Signal, options: Dict[str, object], parent: QWidget):
        super().__init__(options, parent)

        QApplication.instance().aboutToQuit.connect(self.onClose)

        self.server = LiveThread(address, port, self.onConnection)
        self.server.start()
        self.connectionReceived.connect(self.createLivePlayerTab)
        self.queue = Queue()
        self.updateCountSignal = updateCountSignal

    def onConnection(self) -> asyncio.Protocol:
        self.connectionReceived.emit()
        tab = self.queue.get()
        return tab.getProtocol()

    def createLivePlayerTab(self):
        tab = LiveTab(parent=self)
        tab.addIconToTab.connect(self.addIconToTab)
        tab.renameTab.connect(self.renameLivePlayerTab)
        tab.connectionClosed.connect(self.onConnectionClosed)
        self.addTab(tab, "New connection")

        if self.options.get("focusNewTab"):
            self.setCurrentIndex(self.count() - 1)

        self.updateCountSignal.emit()
        self.queue.put(tab)

    def addIconToTab(self, tab: LiveTab):
        index = self.indexOf(tab)
        icon = QFileIconProvider().icon(QFileIconProvider.IconType.Drive)
        self.setTabIcon(index, icon)

    def renameLivePlayerTab(self, tab: LiveTab, name: str):
        index = self.indexOf(tab)
        self.setTabText(index, name)

    def onClose(self):
        self.server.stop()

    def onConnectionClosed(self, tab: LiveTab):
        index = self.indexOf(tab)
        text = self.tabText(index)
        name = text + self.closedTabText
        self.setTabText(index, name)
        self.eventHandler.cleanup()

    def sendKeySequence(self, keys: [Qt.Key]):
        tab: LiveTab = self.currentWidget()

        if tab is not None:
            tab.sendKeySequence(keys)

    def sendText(self, text: str):
        tab: LiveTab = self.currentWidget()

        if tab is not None:
            tab.sendText(text)

    def onTabClosed(self, index: int):
        """
        Gracefully closes the tab by calling the onClose method
        :param index: Index of the closed tab
        """
        super().onTabClosed(index)
        self.updateCountSignal.emit()

    def onTabCloseRequest(self, index: int):
        """
        Prompt the user for validation when the connection is live, then forward call to the parent.
        """
        text = self.tabText(index)

        if not text.endswith(self.closedTabText):
            reply = QMessageBox.question(self, "Confirm close", "Are you sure you want to close a tab with an active connection?", QMessageBox.Yes|QMessageBox.No)
            if reply == QMessageBox.No:
                return

        super().onTabCloseRequest(index)



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Dict, List, Optional

from Crypto.PublicKey import RSA

from pyrdp.enum import NegotiationProtocols, ParserMode
from pyrdp.layer import FastPathLayer, SecurityLayer, TLSSecurityLayer
from pyrdp.parser import createFastPathParser
from pyrdp.pdu import ClientChannelDefinition
from pyrdp.security import RC4CrypterProxy, SecuritySettings
from pyrdp.mitm import MITMConfig


class RDPMITMState:
    """
    State object for the RDP MITM. This is for data that needs to be shared across components.
    """

    def __init__(self, config: MITMConfig, sessionID: str):
        self.requestedProtocols: Optional[NegotiationProtocols] = None
        """The original request protocols"""

        self.config = config
        """The MITM configuration."""

        self.useTLS = False
        """Whether the connection uses TLS or not"""

        self.securitySettings = SecuritySettings()
        """The security settings for the connection"""

        self.channelDefinitions: List[ClientChannelDefinition] = []
        """The channel definitions from the client"""

        self.channelMap: Dict[int, str] = {}
        """Dictionary of channel names to channel IDs"""

        self.rc4RSAKey = RSA.generate(2048)
        """The RSA key for the RC4 key exchange"""

        self.crypters = {
            ParserMode.CLIENT: RC4CrypterProxy(RC4CrypterProxy.Mode.CLIENT),
            ParserMode.SERVER: RC4CrypterProxy(RC4CrypterProxy.Mode.SERVER)
        }
        """Crypters for the client and server side"""

        self.forwardInput = True
        """Whether input from the client should be forwarded to the server"""

        self.forwardOutput = True
        """Whether output from the server should be forwarded to the client"""

        self.loggedIn = False
        """Keep tracks of the client login status"""

        self.inputBuffer = ""
        """Used to store what the client types"""

        self.credentialsCandidate = ""
        """The potential client password"""

        self.shiftPressed = False
        """The current keyboard shift state"""

        self.capsLockOn = False
        """The current keyboard capsLock state"""

        self.ctrlPressed = False
        """The current keybaord ctrl state"""

        self.sessionID = sessionID
        """The current session ID"""

        self.clientIp = None
        """The current client IP address"""

        self.windowSize = None

        self.effectiveTargetHost = self.config.targetHost
        """The host that is currently used as a connection target. It becomes the redirection host when redirection is necessary."""

        self.effectiveTargetPort = self.config.targetPort
        """Port for the effective host"""

        self.ntlmCapture = False
        """Hijack connection from server and capture NTML hash"""

        self.securitySettings.addObserver(self.crypters[ParserMode.CLIENT])
        self.securitySettings.addObserver(self.crypters[ParserMode.SERVER])

    def createSecurityLayer(self, mode: ParserMode, isVirtualChannel: bool) -> SecurityLayer:
        """
        Create a security layer.
        :param mode: the mode of the security layer (client or server)
        :param isVirtualChannel: True if the security layer is for a virtual channel, False if it's for slow-path data.
        """

        if self.useTLS:
            layer = TLSSecurityLayer()
            layer.securityHeaderExpected = not isVirtualChannel
            return layer
        else:
            crypter = self.crypters[mode]
            return SecurityLayer.create(self.securitySettings.encryptionMethod, crypter)

    def createFastPathLayer(self, mode: ParserMode) -> FastPathLayer:
        """
        Create a fast-path layer.
        :param mode: the mode of the layer (client or server)
        """

        parser = createFastPathParser(self.useTLS, self.securitySettings.encryptionMethod, self.crypters[mode], mode)
        return FastPathLayer(parser)

    def canRedirect(self) -> bool:
        return None not in [self.config.redirectionHost, self.config.redirectionPort] and not self.isRedirected()

    def isRedirected(self) -> bool:
        return self.effectiveTargetHost == self.config.redirectionHost and self.effectiveTargetPort == self.config.redirectionPort

    def useRedirectionHost(self):
        self.effectiveTargetHost = self.config.redirectionHost
        self.effectiveTargetPort = self.config.redirectionPort



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO
from typing import Union

from pyrdp.core import Uint16LE, Uint32LE, Uint8
from pyrdp.enum import NegotiationRequestFlags, NegotiationType
from pyrdp.parser.parser import Parser
from pyrdp.pdu import NegotiationFailurePDU, NegotiationRequestPDU, NegotiationResponsePDU


class NegotiationRequestParser(Parser):
    """
    Parser for RDP negotiaton requests (Connection Request payloads).
    """
    def doParse(self, data: bytes) -> NegotiationRequestPDU:
        """
        Parse a negotiation request.
        :param data: the request data.
        """

        cookie = None

        if b"\r\n" in data:
            cookie = data[: data.index(b"\r\n")]
            data = data[data.index(b"\r\n") + 2 :]

        stream = BytesIO(data)

        if len(data) >= 8:
            type = Uint8.unpack(stream)
            requestFlags = Uint8.unpack(stream)
            requestLength = Uint16LE.unpack(stream)
            requestedProtocols = Uint32LE.unpack(stream)

            correlationFlags = None
            correlationID = None

            if requestFlags & NegotiationRequestFlags.CORRELATION_INFO_PRESENT != 0 and len(data) >= 36:
                type = Uint8.unpack(stream)
                correlationFlags = Uint8.unpack(stream)
                correlationLength = Uint16LE.unpack(stream)
                correlationID = stream.read(16)
                stream.read(16)

            return NegotiationRequestPDU(cookie, requestFlags, requestedProtocols, correlationFlags, correlationID)
        else:
            return NegotiationRequestPDU(cookie, None, None, None, None)

    def write(self, pdu):
        """
        Write a negotiation request.
        :param pdu: the request PDU.
        :type pdu: NegotiationRequestPDU
        :return: str
        """
        stream = BytesIO()

        if pdu.cookie is not None:
            stream.write(pdu.cookie + b"\r\n")

        if pdu.flags is not None and pdu.requestedProtocols is not None:
            Uint8.pack(NegotiationType.TYPE_RDP_NEG_REQ, stream)
            Uint8.pack(pdu.flags, stream)
            Uint16LE.pack(8, stream)
            Uint32LE.pack(pdu.requestedProtocols, stream)

            if pdu.correlationFlags is not None and pdu.correlationID is not None:
                Uint8.pack(NegotiationType.TYPE_RDP_CORRELATION_INFO, stream)
                Uint8.pack(pdu.correlationFlags, stream)
                Uint16LE.pack(36, stream)
                stream.write(pdu.correlationID)
                stream.write(b"\x00" * 16)

        return stream.getvalue()


class NegotiationResponseParser(Parser):
    """
    Parser for RDP negotiation responses (Connection Confirm payloads).
    """
    def doParse(self, data: bytes) -> Union[NegotiationResponsePDU, NegotiationFailurePDU]:
        """
        Parse a negotiation response.
        :param data: the response data.
        """
        stream = BytesIO(data)

        if len(data) == 8:
            type = Uint8.unpack(stream)
            flags = Uint8.unpack(stream)
            length = Uint16LE.unpack(stream)

            if type == NegotiationType.TYPE_RDP_NEG_FAILURE:
                failureCode = Uint32LE.unpack(stream)
                return NegotiationFailurePDU(type, flags, failureCode)
            else:
                selectedProtocols = Uint32LE.unpack(stream)
                return NegotiationResponsePDU(type, flags, selectedProtocols)
        else:
            return NegotiationResponsePDU(None, None, None)

    def write(self, pdu):
        """
        Write a negotiation response.
        :param pdu: the response PDU.
        :type pdu: NegotiationResponsePDU
        :return: str
        """
        stream = BytesIO()

        if pdu.flags is not None and pdu.selectedProtocols is not None:
            Uint8.pack(NegotiationType.TYPE_RDP_NEG_RSP, stream)
            Uint8.pack(pdu.flags, stream)
            Uint16LE.pack(8, stream)
            Uint32LE.pack(pdu.selectedProtocols, stream)

        return stream.getvalue()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.pdu.pdu import PDU


class ClientExtraInfo(PDU):
    def __init__(self, clientAddressFamily: int, clientAddress: bytes, clientDir: bytes):
        PDU.__init__(self, b"")
        self.clientAddressFamily = clientAddressFamily
        self.clientAddress = clientAddress
        self.clientDir = clientDir
        self.clientTimeZone = None
        self.clientSessionID = None
        self.performanceFlags = None
        self.autoReconnectCookie = None
        self.dynamicDSTTimeZoneKeyName = None
        self.dynamicDaylightTimeDisabled = None

class ClientInfoPDU(PDU):
    def __init__(self, codePage: int, flags: int, domain: str, username: str, password: str, alternateShell: str, workingDir: str, extraInfo: ClientExtraInfo):
        PDU.__init__(self)
        self.codePage = codePage
        self.flags = flags
        self.domain = domain
        self.username = username
        self.password = password
        self.alternateShell = alternateShell
        self.workingDir = workingDir
        self.extraInfo = extraInfo


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import PlayerPDUType
from pyrdp.layer import FastPathObserver, SlowPathObserver
from pyrdp.pdu import ConfirmActivePDU, InputPDU, UpdatePDU
from pyrdp.pdu.rdp.fastpath import FastPathPDU
from pyrdp.recording.recorder import Recorder


class RecordingFastPathObserver(FastPathObserver):
    def __init__(self, recorder: Recorder, messageType: PlayerPDUType):
        self.recorder = recorder
        self.messageType = messageType
        FastPathObserver.__init__(self)

    def onPDUReceived(self, pdu: FastPathPDU):
        self.recorder.record(pdu, self.messageType)
        FastPathObserver.onPDUReceived(self, pdu)


class RecordingSlowPathObserver(SlowPathObserver):
    def __init__(self, recorder: Recorder):
        SlowPathObserver.__init__(self)
        self.recorder = recorder

    def onPDUReceived(self, pdu):
        if isinstance(pdu, (ConfirmActivePDU, UpdatePDU, InputPDU)):
            self.recorder.record(pdu, PlayerPDUType.SLOW_PATH_PDU)
        SlowPathObserver.onPDUReceived(self, pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.parser import Parser
from pyrdp.pdu import PDU
from pyrdp.pdu.tcp import TCPPDU


class TCPParser(Parser):
    def doParse(self, data: bytes) -> TCPPDU:
        return TCPPDU(data)

    def write(self, pdu: PDU) -> bytes:
        return pdu.payload


#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
Parse Drawing Orders.
"""
import logging
from io import BytesIO

from pyrdp.core.packing import Uint8, Uint16LE
from pyrdp.pdu.rdp.fastpath import FastPathOrdersEvent
from pyrdp.enum.orders import DrawingOrderControlFlags as ControlFlags
from pyrdp.enum.rdp import GlyphSupport
from pyrdp.pdu.rdp.capability import CapabilityType

from .frontend import GdiFrontend

from .secondary import CacheBitmapV1, CacheColorTable, CacheGlyph, CacheBitmapV2, CacheBrush, CacheBitmapV3
from .alternate import CreateOffscreenBitmap, SwitchSurface, CreateNineGridBitmap, \
    StreamBitmapFirst, StreamBitmapNext, GdiPlusFirst, GdiPlusNext, GdiPlusEnd, GdiPlusCacheFirst, \
    GdiPlusCacheNext, GdiPlusCacheEnd, FrameMarker
from .primary import PrimaryContext as Context

LOG = logging.getLogger(__name__)


def _repr(n):
    """Internal method to stringify an order type."""
    r = n.__doc__
    return r if r else 'UNKNOWN (%02x)'.format(n)


class OrdersParser:
    """
    Drawing Order Parser.
    """

    def __init__(self, frontend: GdiFrontend):
        """
        Create a drawing order parser.

        :param GdiFrontend frontend: The frontend that will process GDI messages.
        """

        self.notify: GdiFrontend = frontend
        self.ctx = Context()
        self.glyphLevel: GlyphSupport = GlyphSupport.GLYPH_SUPPORT_NONE

    def onCapabilities(self, caps):
        """Update the parser to take into account the capabilities reported by the client."""

        if CapabilityType.CAPSTYPE_GLYPHCACHE in caps:
            glyphLevel = caps[CapabilityType.CAPSTYPE_GLYPHCACHE].glyphSupportLevel
            self.glyphLevel = GlyphSupport(glyphLevel)

    def parse(self, orders: FastPathOrdersEvent):
        """
        Entrypoint for parsing TS_FP_UPDATE_ORDERS.
        """

        s = BytesIO(orders.payload)

        numberOrders = Uint16LE.unpack(s)
        try:
            for _ in range(numberOrders):
                self._parse_order(s)
        except Exception:
            LOG.warning('Failed to parse drawing order PDU: %s', orders)

        return orders

    def _parse_order(self, s: BytesIO):
        controlFlags = Uint8.unpack(s)

        if not (controlFlags & ControlFlags.TS_STANDARD):
            self._parse_altsec(s, controlFlags)
        elif (controlFlags & ControlFlags.TS_SECONDARY):
            self._parse_secondary(s, controlFlags)
        else:
            self._parse_primary(s, controlFlags)

    # Primary drawing orders.
    # ----------------------------------------------------------------------
    def _parse_primary(self, s: BytesIO, flags: int):

        orderType = self.ctx.update(s, flags)
        self.notify.onBounds(self.ctx.bounds if self.ctx.bounded else None)

        assert orderType >= 0 and orderType < len(_pri)
        _pri[orderType](self, s)

    def _parse_dstblt(self, s: BytesIO):
        """DSTBLT"""
        self.notify.dstBlt(self.ctx.dstBlt.update(s))

    def _parse_patblt(self, s: BytesIO):
        """PATBLT"""
        self.notify.patBlt(self.ctx.patBlt.update(s))

    def _parse_scrblt(self, s: BytesIO):
        """SCRBLT"""
        self.notify.scrBlt(self.ctx.scrBlt.update(s))

    def _parse_draw_nine_grid(self, s: BytesIO):
        """DRAW_NINE_GRID"""
        self.notify.drawNineGrid(self.ctx.drawNineGrid.update(s))

    def _parse_multi_draw_nine_grid(self, s: BytesIO):
        """MULTI_DRAW_NINE_GRID"""
        self.notify.multiDrawNineGrid(self.ctx.multiDrawNineGrid.update(s))

    def _parse_line_to(self, s: BytesIO):
        """LINE_TO"""
        self.notify.lineTo(self.ctx.lineTo.update(s))

    def _parse_opaque_rect(self, s: BytesIO):
        """OPAQUE_RECT"""
        self.notify.opaqueRect(self.ctx.opaqueRect.update(s))

    def _parse_save_bitmap(self, s: BytesIO):
        """SAVE_BITMAP"""
        self.notify.saveBitmap(self.ctx.saveBitmap.update(s))

    def _parse_memblt(self, s: BytesIO):
        """MEMBLT"""
        self.notify.memBlt(self.ctx.memBlt.update(s))

    def _parse_mem3blt(self, s: BytesIO):
        """MEM3BLT"""
        self.notify.mem3Blt(self.ctx.mem3Blt.update(s))

    def _parse_multi_dstblt(self, s: BytesIO):
        """MULTI_DSTBLT"""
        self.notify.multiDstBlt(self.ctx.multiDstBlt.update(s))

    def _parse_multi_patblt(self, s: BytesIO):
        """MULTI_PATBLT"""
        self.notify.multiPatBlt(self.ctx.multiPatBlt.update(s))

    def _parse_multi_scrblt(self, s: BytesIO):
        """MULTI_SCRBLT"""
        self.notify.multiScrBlt(self.ctx.multiScrBlt.update(s))

    def _parse_multi_opaque_rect(self, s: BytesIO):
        """MULTI_OPAQUE_RECT"""
        self.notify.multiOpaqueRect(self.ctx.multiOpaqueRect.update(s))

    def _parse_fast_index(self, s: BytesIO):
        """FAST_INDEX"""
        self.notify.fastIndex(self.ctx.fastIndex.update(s))

    def _parse_polygon_sc(self, s: BytesIO):
        """POLYGON_SC"""
        self.notify.polygonSc(self.ctx.polygonSc.update(s))

    def _parse_polygon_cb(self, s: BytesIO):
        """POLYGON_CB"""
        self.notify.polygonCb(self.ctx.polygonCb.update(s))

    def _parse_polyLine(self, s: BytesIO):
        """POLYLINE"""
        self.notify.polyLine(self.ctx.polyLine.update(s))

    def _parse_fast_glyph(self, s: BytesIO):
        """FAST_GLYPH"""
        self.notify.fastGlyph(self.ctx.fastGlyph.update(s))

    def _parse_ellipse_sc(self, s: BytesIO):
        """ELLIPSE_SC"""
        self.notify.ellipseSc(self.ctx.ellipseSc.update(s))

    def _parse_ellipse_cb(self, s: BytesIO):
        """ELLIPSE_CB"""
        self.notify.ellipseCb(self.ctx.ellipseCb.update(s))

    def _parse_glyph_index(self, s: BytesIO):
        """GLYPH_INDEX"""
        self.notify.glyphIndex(self.ctx.glyphIndex.update(s))

    # Secondary drawing orders.
    # ----------------------------------------------------------------------
    def _parse_secondary(self, s: BytesIO, flags: int):
        Uint16LE.unpack(s)  # orderLength (unused)
        extraFlags = Uint16LE.unpack(s)
        orderType = Uint8.unpack(s)

        assert orderType >= 0 and orderType < len(_sec)
        _sec[orderType](self, s, orderType, extraFlags)

    def _parse_cache_bitmap_v1(self, s: BytesIO, orderType: int, flags: int):
        """CACHE_BITMAP_V1"""
        self.notify.cacheBitmapV1(CacheBitmapV1.parse(s, orderType, flags))

    def _parse_cache_color_table(self, s: BytesIO, orderType: int, flags: int):
        """CACHE_COLOR_TABLE"""
        self.notify.cacheColorTable(CacheColorTable.parse(s, orderType, flags))

    def _parse_cache_glyph(self, s: BytesIO, orderType: int, flags: int):
        """CACHE_GLYPH"""
        if self.glyphLevel == GlyphSupport.GLYPH_SUPPORT_NONE:
            LOG.warn("Received CACHE_GLYPH but the client reported it doesn't support it!")
            # Ignore it.
        else:
            self.notify.cacheGlyph(CacheGlyph.parse(s, flags, self.glyphLevel))

    def _parse_cache_bitmap_v2(self, s: BytesIO, orderType: int, flags: int):
        """CACHE_BITMAP_V2"""
        self.notify.cacheBitmapV2(CacheBitmapV2.parse(s, orderType, flags))

    def _parse_cache_brush(self, s: BytesIO, orderType: int, flags: int):
        """CACHE_BRUSH"""
        self.notify.cacheBrush(CacheBrush.parse(s))

    def _parse_cache_bitmap_v3(self, s: BytesIO, orderType: int, flags: int):
        """CACHE_BITMAP_V3"""
        self.notify.cacheBitmapV3(CacheBitmapV3.parse(s, flags))

    # Alternate secondary drawing orders.
    # ----------------------------------------------------------------------
    def _parse_altsec(self, s: BytesIO, flags: int):
        orderType = flags >> 2

        assert orderType >= 0 and orderType < len(_alt)

        _alt[orderType](self, s)

    def _parse_create_offscreen_bitmap(self, s: BytesIO):
        """CREATE_OFFSCREEN_BITMAP"""
        self.notify.createOffscreenBitmap(CreateOffscreenBitmap.parse(s))

    def _parse_switch_surface(self, s: BytesIO):
        """SWITCH_SURFACE"""
        self.notify.switchSurface(SwitchSurface.parse(s))

    def _parse_create_nine_grid_bitmap(self, s: BytesIO):
        """CREATE_NINEGRID_BITMAP"""
        self.notify.createNineGridBitmap(CreateNineGridBitmap.parse(s))

    def _parse_stream_bitmap_first(self, s: BytesIO):
        """STREAM_BITMAP_FIRST"""
        self.notify.streamBitmapFirst(StreamBitmapFirst.parse(s))

    def _parse_stream_bitmap_next(self, s: BytesIO):
        """STREAM_BITMAP_NEXT"""
        self.notify.streamBitmapNext(StreamBitmapNext.parse(s))

    def _parse_gdiplus_first(self, s: BytesIO):
        """GDIPLUS_FIRST"""
        self.notify.gdiPlusFirst(GdiPlusFirst.parse(s))

    def _parse_gdiplus_next(self, s: BytesIO):
        """GDIPLUS_NEXT"""
        self.notify.drawGdiPlusNext(GdiPlusNext.parse(s))

    def _parse_gdiplus_end(self, s: BytesIO):
        """GDIPLUS_END"""
        self.notify.drawGdiPlusEnd(GdiPlusEnd.parse(s))

    def _parse_gdiplus_cache_first(self, s: BytesIO):
        """GDIPLUS_CACHE_FIRST"""
        self.notify.drawGdiPlusCacheFirst(GdiPlusCacheFirst.parse(s))

    def _parse_gdiplus_cache_next(self, s: BytesIO):
        """GDIPLUS_CACHE_NEXT"""
        self.notify.drawGdiPlusCacheNext(GdiPlusCacheNext.parse(s))

    def _parse_gdiplus_cache_end(self, s: BytesIO):
        """GDIPLUS_CACHE_END"""
        self.notify.drawGdiPlusCacheEnd(GdiPlusCacheEnd.parse(s))

    def _parse_window(self, s: BytesIO):
        """WINDOW"""
        # This is specified in MS-RDPERP for seamless applications.
        LOG.debug('WINDOW is not supported yet.')

    def _parse_compdesk_first(self, s: BytesIO):
        """COMPDESK"""
        LOG.debug('COMPDESK is not supported yet.')

    def _parse_frame_marker(self, s: BytesIO):
        """FRAME_MARKER"""
        self.notify.frameMarker(FrameMarker.parse(s))


# Parser Lookup Tables
_pri = [
    OrdersParser._parse_dstblt,                # 0x00
    OrdersParser._parse_patblt,                # 0x01
    OrdersParser._parse_scrblt,                # 0x02
    None,                                      # 0x03
    None,                                      # 0x04
    None,                                      # 0x05
    None,                                      # 0x06
    OrdersParser._parse_draw_nine_grid,        # 0x07
    OrdersParser._parse_multi_draw_nine_grid,  # 0x08
    OrdersParser._parse_line_to,               # 0x09
    OrdersParser._parse_opaque_rect,           # 0x0A
    OrdersParser._parse_save_bitmap,           # 0x0B
    None,                                      # 0x0C
    OrdersParser._parse_memblt,                # 0x0D
    OrdersParser._parse_mem3blt,               # 0x0E
    OrdersParser._parse_multi_dstblt,          # 0x0F
    OrdersParser._parse_multi_patblt,          # 0x10
    OrdersParser._parse_multi_scrblt,          # 0x11
    OrdersParser._parse_multi_opaque_rect,     # 0x12
    OrdersParser._parse_fast_index,            # 0x13
    OrdersParser._parse_polygon_sc,            # 0x14
    OrdersParser._parse_polygon_cb,            # 0x15
    OrdersParser._parse_polyLine,              # 0x16
    None,                                      # 0x17
    OrdersParser._parse_fast_glyph,            # 0x18
    OrdersParser._parse_ellipse_sc,            # 0x19
    OrdersParser._parse_ellipse_cb,            # 0x1A
    OrdersParser._parse_glyph_index,           # 0x1B
]

_sec = [
    OrdersParser._parse_cache_bitmap_v1,    # 0x00 : Uncompressed
    OrdersParser._parse_cache_color_table,  # 0x01
    OrdersParser._parse_cache_bitmap_v1,    # 0x02 : Compressed
    OrdersParser._parse_cache_glyph,        # 0x03
    OrdersParser._parse_cache_bitmap_v2,    # 0x04 : Uncompresed
    OrdersParser._parse_cache_bitmap_v2,    # 0x05 : Compressed
    None,                                   # 0x06
    OrdersParser._parse_cache_brush,        # 0x07
    OrdersParser._parse_cache_bitmap_v3,    # 0x08
]

_alt = [
    OrdersParser._parse_switch_surface,           # 0x00
    OrdersParser._parse_create_offscreen_bitmap,  # 0x01
    OrdersParser._parse_stream_bitmap_first,      # 0x02
    OrdersParser._parse_stream_bitmap_next,       # 0x03
    OrdersParser._parse_create_nine_grid_bitmap,  # 0x04
    OrdersParser._parse_gdiplus_first,            # 0x05
    OrdersParser._parse_gdiplus_next,             # 0x06
    OrdersParser._parse_gdiplus_end,              # 0x07
    OrdersParser._parse_gdiplus_cache_first,      # 0x08
    OrdersParser._parse_gdiplus_cache_next,       # 0x09
    OrdersParser._parse_gdiplus_cache_end,        # 0x0A
    OrdersParser._parse_window,                   # 0x0B
    OrdersParser._parse_compdesk_first,           # 0x0C
    OrdersParser._parse_frame_marker,             # 0x0D
]



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.recording.observer import RecordingFastPathObserver, RecordingSlowPathObserver
from pyrdp.recording.recorder import FileLayer, Recorder



#
# This file is part of the PyRDP project.
# Copyright (C) 2020-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import CapabilityType
from pyrdp.pdu import PlayerPDU
from pyrdp.player.ImageHandler import ImageHandler
from pyrdp.player.RenderingEventHandler import RenderingEventHandler

import logging

import av
import qimage2ndarray
from PySide6.QtGui import QImage, QPainter, QColor


class MP4Image(ImageHandler):
    """A QRemoteDesktop Mock."""
    def __init__(self):
        self.buffer: QImage = None

    def notifyImage(self, x: int, y: int, img: QImage, width: int, height: int):
        p = QPainter(self.buffer)
        p.drawImage(x, y, img, 0, 0, width, height)

    def resize(self, width: int, height: int):
        self.buffer = QImage(width, height, QImage.Format_ARGB32_Premultiplied)

    def update(self):
        pass

    @property
    def screen(self) -> QImage:
        return self.buffer


class MP4EventHandler(RenderingEventHandler):

    def __init__(self, filename: str, fps=30, progress=None):
        """
        Construct an event handler that outputs to an Mp4 file.

        :param filename: The output file to write to.
        :param fps: The frame rate (30 recommended).
        :param progress: An optional callback (sig: `() -> ()`) whenever a frame is muxed.
        """
        self.filename = filename
        # The movflags puts the encoder in an MP4 Streaming Format. This has two benefits:
        # - recover partial videos in case of a pyrdp-convert crash
        # - reduce memory consumption (especially for long captures)
        # See: https://ffmpeg.org/ffmpeg-formats.html#mov_002c-mp4_002c-ismv
        self.mp4 = f = av.open(filename, 'w', options={'movflags': 'frag_keyframe+empty_moov'})
        self.stream = f.add_stream('h264', rate=fps)
        # TODO: this undocumented PyAV stream feature needs to be properly investigated
        #       we could probably batch the encoding of several frames and benefit from threads
        #       but trying this as-is lead to no gains
        #       (actually a degradation but that could be statistically irrelevant)
        #self.stream.thread_count = 4
        self.stream.pix_fmt = 'yuv420p'
        self.progress = progress
        self.scale = False
        self.mouse = (0, 0)
        self.fps = fps
        self.delta = 1000 // fps  # ms per frame
        self.log = logging.getLogger(__name__)
        self.log.info('Begin MP4 export to %s: %d FPS', filename, fps)
        self.timestamp = self.prevTimestamp = None

        super().__init__(MP4Image())

    def onPDUReceived(self, pdu: PlayerPDU):
        super().onPDUReceived(pdu)

        # Make sure the rendering surface has been created.
        if self.imageHandler.screen is None:
            return

        ts = pdu.timestamp
        self.timestamp = ts

        if self.prevTimestamp is None:
            dt = self.delta
        else:
            dt = self.timestamp - self.prevTimestamp  # ms
        nframes = (dt // self.delta)
        if nframes > 0:
            for _ in range(nframes):
                self.writeFrame()
            self.prevTimestamp = ts
            self.log.debug('Rendered %d still frame(s)', nframes)

    def cleanup(self):
        # Add one second worth of padding so that the video doesn't end too abruptly.
        for _ in range(self.fps):
            self.writeFrame()

        self.log.info('Flushing to disk: %s', self.filename)
        for pkt in self.stream.encode():
            if self.progress:
                self.progress()
            self.mp4.mux(pkt)
        self.log.info('Export completed.')
        self.mp4.close()

    def onMousePosition(self, x, y):
        self.mouse = (x, y)
        super().onMousePosition(x, y)

    def onCapabilities(self, caps):
        bmp = caps[CapabilityType.CAPSTYPE_BITMAP]
        (w, h) = (bmp.desktopWidth, bmp.desktopHeight)
        self.imageHandler.resize(w, h)

        if w % 2 != 0:
            self.scale = True
            w += 1
        if h % 2 != 0:
            self.scale = True
            h += 1

        self.stream.width = w
        self.stream.height = h

        super().onCapabilities(caps)

    def onFinishRender(self):
        # When the screen is updated, always write a frame.
        self.prevTimestamp = self.timestamp
        self.writeFrame()

    def writeFrame(self):
        w = self.stream.width
        h = self.stream.height
        surface = self.imageHandler.screen.scaled(w, h) if self.scale else self.imageHandler.screen.copy()

        # Draw the mouse pointer. Render mouse clicks?
        p = QPainter(surface)
        p.setBrush(QColor.fromRgb(255, 255, 0, 180))
        (x, y) = self.mouse
        p.drawEllipse(x, y, 5, 5)
        p.end()

        # Output frame.
        frame = av.VideoFrame.from_ndarray(qimage2ndarray.rgb_view(surface))
        for packet in self.stream.encode(frame):
            if self.progress:
                self.progress()
            self.mp4.mux(packet)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO
from typing import Tuple

from pyrdp.core import Uint16BE, Uint16LE, Uint8
from pyrdp.enum import X224PDUType
from pyrdp.exceptions import ParsingError, UnknownPDUTypeError
from pyrdp.parser.parser import Parser
from pyrdp.pdu import X224ConnectionConfirmPDU, X224ConnectionRequestPDU, X224DataPDU, X224DisconnectRequestPDU, \
    X224ErrorPDU, X224PDU


class X224Parser(Parser):
    """
    Parser to read and write X224 (COTP) PDUs.
    """

    def __init__(self):
        super().__init__()
        self.parsers = {
            X224PDUType.X224_TPDU_CONNECTION_REQUEST: self.parseConnectionRequest,
            X224PDUType.X224_TPDU_CONNECTION_CONFIRM: self.parseConnectionConfirm,
            X224PDUType.X224_TPDU_DISCONNECT_REQUEST: self.parseDisconnectRequest,
            X224PDUType.X224_TPDU_DATA: self.parseData,
            X224PDUType.X224_TPDU_ERROR: self.parseError,
        }

        self.writers = {
            X224PDUType.X224_TPDU_CONNECTION_REQUEST: self.writeConnectionRequest,
            X224PDUType.X224_TPDU_CONNECTION_CONFIRM: self.writeConnectionConfirm,
            X224PDUType.X224_TPDU_DISCONNECT_REQUEST: self.writeDisconnectRequest,
            X224PDUType.X224_TPDU_DATA: self.writeData,
            X224PDUType.X224_TPDU_ERROR: self.writeError,
        }

    def doParse(self, data: bytes) -> X224PDU:
        """
        Read the byte stream and return a corresponding X224PDU
        """
        length = Uint8.unpack(data[0])
        header = Uint8.unpack(data[1]) >> 4

        if header in list(X224PDUType):
            header = X224PDUType(header)

        if length < 2:
            raise ParsingError("Invalid X224 length indicator: indicator = %d, expected at least 2 bytes" % length)
        if len(data) < length:
            raise ParsingError("Invalid X224 length indicator: indicator = %d, length = %d" % (length, len(data)))

        if header not in self.parsers:
            raise UnknownPDUTypeError("Trying to parse unknown X224 PDU type: %s" % (header if header in X224PDUType else hex(header)), header)

        return self.parsers[header](data, length)

    def parseConnectionPDU(self, data: bytes, length: int, name: str) -> Tuple[int, int, int, bytes]:
        """
        Parse the provided data to extract common information contained in Connection Request, Connection Confirm and Disconnect Request PDUs.
        :param data: bytes to parse.
        :param length: Length of the Connection PDU.
        :param name: For debugging purposes: the name of the connection PDU (like "Connection Request").
        :return: A tuple of the information we find in both connection PDUs: (source, destination, options, payload)
        """

        if length < 6:
            raise ParsingError("Invalid X244 %s length indicator: indicator = %d, expected at least 6 bytes" % (name, 6))

        destination = Uint16BE.unpack(data[2 : 4])
        source = Uint16BE.unpack(data[4 : 6])
        options = Uint8.unpack(data[6])
        payload = data[7 :]

        if len(payload) != length - 6:
            raise ParsingError("Invalid X224 %s payload length: expected = %d, length = %d" % (name, length - 6, len(payload)))

        return source, destination, options, payload

    def parseConnectionRequest(self, data: bytes, length: int) -> X224ConnectionRequestPDU:
        """
        Parse a ConnectionRequest PDU from the raw bytes
        :param data: bytes to parse.
        :param length: The length in bytes of the Connection Request PDU.
        """
        credit = Uint8.unpack(data[1]) & 0xf
        destination, source, options, payload = self.parseConnectionPDU(data, length, "Connection Request")
        return X224ConnectionRequestPDU(credit, destination, source, options, payload)

    def parseConnectionConfirm(self, data: bytes, length: int) -> X224ConnectionConfirmPDU:
        """
        Parse a ConnectionConfirm PDU from the raw bytes
        :param data: bytes to parse.
        :param length: The length in bytes of the Connection Confirm PDU.
        """
        credit = Uint8.unpack(data[1]) & 0xf
        destination, source, options, payload = self.parseConnectionPDU(data, length, "Connection Confirm")
        return X224ConnectionConfirmPDU(credit, destination, source, options, payload)

    def parseDisconnectRequest(self, data: bytes, length: int) -> X224DisconnectRequestPDU:
        """
        Parse a DisconnectRequest PDU from the raw bytes
        :param data: bytes to parse.
        :param length: The length in bytes of the Disconnect Request PDU.
        """
        destination, source, reason, payload = self.parseConnectionPDU(data, length, "Disconnect Request")
        return X224DisconnectRequestPDU(destination, source, reason, payload)

    def parseData(self, data: bytes, length: int) -> X224DataPDU:
        """
        Parse a Data PDU from the raw bytes
        :param data: bytes to parse.
        :param length: The length in bytes of the Data PDU.
        """

        if length != 2:
            raise ParsingError("Invalid X224 Data PDU length indicator: expected = 2, indicator = %d" % length)

        code = Uint8.unpack(data[1]) & 0xf
        sequence = Uint8.unpack(data[2])
        payload = data[3 :]

        return X224DataPDU(code & 1 == 1, sequence & 0x80 == 0x80, payload)

    def parseError(self, data: bytes, length: int) -> X224ErrorPDU:
        """
        Parse a Error PDU from the raw bytes
        :param data: bytes to parse.
        :param length: The length in bytes of the Error PDU.
        """

        if length < 4:
            raise ParsingError("Invalid X224 Error PDU length indicator: indicator = %d, expected at least 4 bytes")

        destination = Uint16BE.unpack(data[2 : 4])
        cause = Uint8.unpack(data[4])
        payload = data[5 :]

        if len(payload) != length - 4:
            raise ParsingError("Invalid X224 Error PDU payload length: expected = %d, length = %d" % (length - 4, len(payload)))

        return X224ErrorPDU(destination, cause, payload)

    def write(self, pdu: X224PDU) -> bytes:
        """
        Encode the provided X224 pdu into a byte stream.
        :return: The bytes to send to the previous layer
        """

        stream = BytesIO()


        if pdu.header == X224PDUType.X224_TPDU_DATA:
            length = 2
        elif pdu.header in [X224PDUType.X224_TPDU_CONNECTION_REQUEST, X224PDUType.X224_TPDU_CONNECTION_CONFIRM, X224PDUType.X224_TPDU_DISCONNECT_REQUEST]:
            length = len(pdu.payload) + 6
        elif pdu.header == X224PDUType.X224_TPDU_ERROR:
            length = len(pdu.payload) + 4
        else:
            raise UnknownPDUTypeError("Trying to write unknown X224 PDU type: %s" % (pdu.header if pdu.header in X224PDUType else hex(pdu.header)), pdu.header)

        stream.write(Uint8.pack(length))
        self.writers[pdu.header](stream, pdu)
        stream.write(pdu.payload)
        return stream.getvalue()

    def writeConnectionPDU(self, stream: BytesIO, header: X224PDUType, destination: int, source: int, options: int):
        """
        Write a connection PDU (connectionRequest/connectionConfirm/disconnectRequest) in the provided byte stream.
        """
        stream.write(Uint8.pack(header))
        stream.write(Uint16BE.pack(destination))
        stream.write(Uint16BE.pack(source))
        stream.write(Uint8.pack(options))

    def writeConnectionRequest(self, stream: BytesIO, pdu: X224ConnectionRequestPDU):
        """
        Write a connection request PDU onto the provided stream
        """
        header = (pdu.header << 4) | (pdu.credit & 0xf)
        self.writeConnectionPDU(stream, header, pdu.destination, pdu.source, pdu.options)

    def writeConnectionConfirm(self, stream: BytesIO, pdu: X224ConnectionConfirmPDU):
        """
        Write a connection confirm PDU onto the provided stream
        """
        header = (pdu.header << 4) | (pdu.credit & 0xf)
        self.writeConnectionPDU(stream, header, pdu.destination, pdu.source, pdu.options)

    def writeDisconnectRequest(self, stream: BytesIO, pdu: X224DisconnectRequestPDU):
        """
        Write a disconnect request PDU onto the provided stream
        """
        self.writeConnectionPDU(stream, pdu.header, pdu.destination, pdu.source, pdu.reason)

    def writeData(self, stream: BytesIO, pdu: X224DataPDU):
        """
        Write a Data PDU onto the provided stream
        """
        header = (pdu.header << 4) | int(pdu.roa)
        stream.write(Uint8.pack(header))
        stream.write(Uint8.pack(int(pdu.eot) << 7))

    def writeError(self, stream: BytesIO, pdu: X224ErrorPDU):
        """
        Write an error PDU onto the provided stream
        """
        stream.write(Uint8.pack(pdu.header))
        stream.write(Uint16LE.pack(pdu.destination))
        stream.write(Uint8.pack(pdu.cause))



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from logging import LoggerAdapter

from pyrdp.layer import TwistedTCPLayer
from pyrdp.logging.StatCounter import StatCounter
from pyrdp.mitm.state import RDPMITMState
from pyrdp.pdu.player import PlayerConnectionClosePDU
from pyrdp.recording import Recorder


class TCPMITM:
    """
    MITM component for the TCP layer.
    """

    def __init__(self, client: TwistedTCPLayer, server: TwistedTCPLayer, attacker: TwistedTCPLayer, log: LoggerAdapter,
                 state: RDPMITMState, recorder: Recorder, statCounter: StatCounter):
        """
        :param client: TCP layer for the client side
        :param server: TCP layer for the server side
        :param attacker: TCP layer for the attacker side
        :param log: logger for this component
        :param recorder: recorder for this connection
        """

        self.statCounter = statCounter
        # To keep track of useful statistics for the connection.
        self.client = client
        self.server = None
        self.attacker = attacker
        self.log = log
        self.state = state
        self.recorder = recorder

        # Allows a lower layer to raise error tagged with the correct sessionID
        self.client.log = log

        self.clientObserver = self.client.createObserver(
            onConnection = self.onClientConnection,
            onDisconnection = self.onClientDisconnection,
        )

        self.attacker.createObserver(
            onConnection = self.onAttackerConnection,
            onDisconnection = self.onAttackerDisconnection,
        )

        self.serverObserver = None
        self.setServer(server)

    def setServer(self, server: TwistedTCPLayer):
        if self.server is not None:
            self.server.removeObserver(self.serverObserver)
            self.server.disconnect(True)

        self.server = server
        self.server.log = self.log
        self.serverObserver = self.server.createObserver(
            onConnection=self.onServerConnection,
            onDisconnection=self.onServerDisconnection,
        )

    def detach(self):
        """
        Remove the observers from the layers.
        """

        self.client.removeObserver(self.clientObserver)
        self.server.removeObserver(self.serverObserver)

    def onClientConnection(self):
        """
        Log the fact that a new client has connected.
        """
        
        # Statistics
        self.statCounter.start()        

        ip = self.client.transport.client[0]
        port = self.client.transport.client[1]
        self.state.clientIp = ip
        self.log.extra['clientIp'] = ip
        self.log.info("New client connected from %(clientIp)s:%(clientPort)i",
                      {"clientIp": ip, "clientPort": port})

    def onClientDisconnection(self, reason):
        """
        Disconnect all the parts of the connection.
        :param reason: reason for disconnection
        """

        self.statCounter.stop()
        self.recordConnectionClose()
        self.log.info("Client connection closed. %(reason)s", {"reason": reason.value})
        if self.recorder.recordFilename:
            self.statCounter.logReport(self.log, {"replayFilename":
                                                  self.recorder.recordFilename})
        else:
            self.statCounter.logReport(self.log)

        self.recorder.finalize()
        self.server.disconnect(True)
        self.state.clientIp = None

        # For the attacker, we want to make sure we don't abort the connection to make sure that the close event is sent
        self.attacker.disconnect()
        self.detach()

    def onServerConnection(self):
        """
        Log the fact that a connection to the server was established.
        """
        self.log.info("Server connected")

    def onServerDisconnection(self, reason):
        """
        Disconnect all the parts of the connection.
        :param reason: reason for disconnection
        """

        self.recordConnectionClose()
        self.recorder.finalize()
        self.log.info("Server connection closed. %(reason)s", {"reason": reason.value})
        self.client.disconnect(True)

        # For the attacker, we want to make sure we don't abort the connection to make sure that the close event is sent
        self.attacker.disconnect()
        self.detach()

    def onAttackerConnection(self):
        """
        Log the fact that a connection to the attacker was established.
        """
        self.log.info("Attacker connected")

    def onAttackerDisconnection(self, reason):
        """
        Log the disconnection from the attacker side.
        """
        self.state.forwardInput = True
        self.state.forwardOutput = True
        self.log.info("Attacker connection closed. %(reason)s", {"reason": reason.value})

    def recordConnectionClose(self):
        pdu = PlayerConnectionClosePDU(self.recorder.getCurrentTimeStamp())
        self.recorder.record(pdu, pdu.header)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import VirtualChannelPDUFlag
from pyrdp.layer.layer import IntermediateLayer
from pyrdp.logging import log
from pyrdp.parser import VirtualChannelParser
from pyrdp.pdu import PDU, VirtualChannelPDU


class VirtualChannelLayer(IntermediateLayer):
    """
    Layer that handles the virtual channel layer of the RDP protocol:
    https://msdn.microsoft.com/en-us/library/cc240548.aspx
    """

    def __init__(self, parser = VirtualChannelParser(), activateShowProtocolFlag = True):
        """
        :param activateShowProtocolFlag: True if the channelFlagShowProtocol must be set (depends on virtual channels)
        """
        super().__init__(parser)
        self.activateShowProtocolFlag = activateShowProtocolFlag
        self.pduBuffer = b""

    def recv(self, data: bytes):
        virtualChannelPDU = self.mainParser.parse(data)

        if virtualChannelPDU.flags & VirtualChannelPDUFlag.CHANNEL_PACKET_COMPRESSED != 0:
            log.error("Compression flag is set on virtual channel data, it is NOT handled, crash will most likely occur.")

        flags = virtualChannelPDU.flags
        if flags & VirtualChannelPDUFlag.CHANNEL_FLAG_FIRST:
            self.pduBuffer = virtualChannelPDU.payload
        else:
            self.pduBuffer += virtualChannelPDU.payload

        if flags & VirtualChannelPDUFlag.CHANNEL_FLAG_LAST:
            # Reassembly done, change the payload of the virtualChannelPDU for processing by the observer.
            virtualChannelPDU.payload = self.pduBuffer
            self.pduReceived(virtualChannelPDU)

    def sendBytes(self, payload: bytes):
        """
        Send data to the previous layer by encapsulating it in a VirtualChannelPDU and fragmenting it if necessary.
        """
        flags: VirtualChannelPDUFlag = 0

        if self.activateShowProtocolFlag:
            flags |= VirtualChannelPDUFlag.CHANNEL_FLAG_SHOW_PROTOCOL

        virtualChannelPDU = VirtualChannelPDU(flags, payload)
        rawVirtualChannelPDUsList = self.mainParser.write(virtualChannelPDU)

        # Since a virtualChannelPDU may need to be sent using several packets
        for data in rawVirtualChannelPDUsList:
            self.previous.sendBytes(data)

    def shouldForward(self, pdu: PDU) -> bool:
        return True


#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.layer import RawLayer
from pyrdp.logging import StatCounter
from pyrdp.logging.StatCounter import STAT
from pyrdp.pdu import PDU


class VirtualChannelMITM:
    """
    Generic MITM component for any virtual channel.
    """

    def __init__(self, client: RawLayer, server: RawLayer, statCounter: StatCounter):
        """
        :param client: layer for the client side
        :param server: layer for the server side
        """

        self.client = client
        self.server = server
        self.statCounter = statCounter

        self.client.createObserver(
            onPDUReceived = self.onClientPDUReceived
        )

        self.server.createObserver(
            onPDUReceived = self.onServerPDUReceived
        )

    def onClientPDUReceived(self, pdu: PDU):
        """
        Forward the PDU to the server.
        :param pdu: the PDU that was received
        """

        self.statCounter.increment(STAT.VIRTUAL_CHANNEL_INPUT)
        self.server.sendPDU(pdu)

    def onServerPDUReceived(self, pdu: PDU):
        """
        Forward the PDU to the client.
        :param pdu: the PDU that was received
        """

        self.statCounter.increment(STAT.VIRTUAL_CHANNEL, STAT.VIRTUAL_CHANNEL_OUTPUT)
        self.client.sendPDU(pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum


class MCSPDUType(IntEnum):
    """
    MCS PDU Headers
    """
    # Connection PDU headers
    CONNECT_INITIAL = 0x65
    CONNECT_RESPONSE = 0x66

    # Domain PDU headers
    ERECT_DOMAIN_REQUEST = 1
    DISCONNECT_PROVIDER_ULTIMATUM = 8
    ATTACH_USER_REQUEST = 10
    ATTACH_USER_CONFIRM = 11
    CHANNEL_JOIN_REQUEST = 14
    CHANNEL_JOIN_CONFIRM = 15
    SEND_DATA_REQUEST = 25
    SEND_DATA_INDICATION = 26


class MCSResult(IntEnum):
    """
    MCS Result enumeration.
    """
    RT_SUCCESSFUL = 0
    RT_UNSPECIFIED_FAILURE = 14
    RT_USER_REJECTED = 15


class MCSChannelID(IntEnum):
    """
    MCS Channel ID definitions.
    """
    USERCHANNEL_BASE = 1001


class MCSChannelName:
    IO = "I/O"
    CLIPBOARD = "cliprdr"
    DEVICE_REDIRECTION = "rdpdr"
    DYNAMIC_CHANNEL = "drdynvc"


#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO
from typing import List

from pyrdp.core import Uint16LE, Uint32LE, Uint8
from pyrdp.enum import CapabilityType, ErrorInfo, SlowPathDataType, SlowPathPDUType
from pyrdp.parser.parser import Parser
from pyrdp.parser.rdp.input import SlowPathInputParser
from pyrdp.parser.rdp.pointer import PointerEventParser
from pyrdp.pdu import BitmapCapability, Capability, ConfirmActivePDU, ControlPDU, DemandActivePDU, GeneralCapability, \
    GlyphCacheCapability, InputPDU, MultifragmentUpdateCapability, OffscreenBitmapCacheCapability, OrderCapability, \
    PDU, PlaySoundPDU, PointerCapability, PointerPDU, SetErrorInfoPDU, ShareControlHeader, ShareDataHeader, \
    SlowPathPDU, SlowPathUnparsedPDU, SuppressOutputPDU, SynchronizePDU, UpdatePDU, VirtualChannelCapability, \
    PersistentCacheKeysPDU
from pyrdp.pdu.rdp.capability import SurfaceCommandsCapability


class SlowPathParser(Parser):
    def __init__(self):
        super().__init__()
        self.parsers = {
            SlowPathPDUType.DEMAND_ACTIVE_PDU: self.parseDemandActive,
            SlowPathPDUType.CONFIRM_ACTIVE_PDU: self.parseConfirmActive,
            SlowPathPDUType.DATA_PDU: self.parseData,
        }

        self.dataParsers = {
            SlowPathDataType.PDUTYPE2_SET_ERROR_INFO_PDU: self.parseError,
            SlowPathDataType.PDUTYPE2_SYNCHRONIZE: self.parseSynchronize,
            SlowPathDataType.PDUTYPE2_CONTROL: self.parseControl,
            SlowPathDataType.PDUTYPE2_INPUT: self.parseInput,
            # RDPDataPDUSubtype.PDUTYPE2_POINTER: self.parsePointer,
            SlowPathDataType.PDUTYPE2_PLAY_SOUND: self.parsePlaySound,
            SlowPathDataType.PDUTYPE2_SUPPRESS_OUTPUT: self.parseSuppressOutput,
            SlowPathDataType.PDUTYPE2_UPDATE: self.parseUpdate,
            SlowPathDataType.PDUTYPE2_BITMAPCACHE_PERSISTENT_LIST: self.parsePersistentCacheKeys,
        }

        self.dataWriters = {
            SlowPathDataType.PDUTYPE2_SET_ERROR_INFO_PDU: self.writeError,
            SlowPathDataType.PDUTYPE2_SYNCHRONIZE: self.writeSynchronize,
            SlowPathDataType.PDUTYPE2_CONTROL: self.writeControl,
            SlowPathDataType.PDUTYPE2_INPUT: self.writeInput,
            # RDPDataPDUSubtype.PDUTYPE2_POINTER: self.writePointer,
            SlowPathDataType.PDUTYPE2_PLAY_SOUND: self.writePlaySound,
            SlowPathDataType.PDUTYPE2_SUPPRESS_OUTPUT: self.writeSuppressOutput,
            SlowPathDataType.PDUTYPE2_UPDATE: self.writeUpdate,
            SlowPathDataType.PDUTYPE2_BITMAPCACHE_PERSISTENT_LIST: self.writePersistentCacheKeys,
        }

    def doParse(self, data: bytes) -> PDU:
        """
        Decode a data PDU from bytes.
        :return: an instance of an RDP Data PDU class.
        """
        stream = BytesIO(data)
        header = self.parseShareControlHeader(stream)

        if header.pduType not in self.parsers:
            return SlowPathUnparsedPDU(header, stream.read())

        return self.parsers[header.pduType](stream, header)

    def parseData(self, stream: BytesIO, header):
        header = self.parseShareDataHeader(stream, header)

        if header.subtype not in self.dataParsers:
            return SlowPathUnparsedPDU(header, stream.read())

        return self.dataParsers[header.subtype](stream, header)

    def write(self, pdu: SlowPathPDU) -> bytes:
        """
        Encode an RDP Data PDU instance to bytes.
        """
        stream = BytesIO()
        substream = BytesIO()

        if isinstance(pdu, DemandActivePDU):
            headerWriter = self.writeShareControlHeader
            self.writeDemandActive(substream, pdu)
        elif isinstance(pdu, ConfirmActivePDU):
            headerWriter = self.writeShareControlHeader
            self.writeConfirmActive(substream, pdu)
        elif pdu.header.pduType == SlowPathPDUType.DATA_PDU:
            headerWriter = self.writeShareDataHeader
            self.writeData(substream, pdu)
        else:
            headerWriter = self.writeShareControlHeader
            substream.write(pdu.payload)

        substream = substream.getvalue()
        headerWriter(stream, pdu.header, len(substream))
        stream.write(substream)
        return stream.getvalue()

    def writeData(self, stream: BytesIO, pdu):
        if pdu.header.subtype not in self.dataWriters:
            stream.write(pdu.payload)
        else:
            self.dataWriters[pdu.header.subtype](stream, pdu)

    def parseShareControlHeader(self, stream: BytesIO):
        Uint16LE.unpack(stream)  # length (unused)
        pduType = Uint16LE.unpack(stream)
        source = Uint16LE.unpack(stream)
        return ShareControlHeader(SlowPathPDUType(pduType & 0xf), (pduType >> 4), source)

    def writeShareControlHeader(self, stream: BytesIO, header: ShareControlHeader, dataLength: int):
        pduType = (header.pduType.value & 0xf) | (header.version << 4)
        stream.write(Uint16LE.pack(dataLength + 6))
        stream.write(Uint16LE.pack(pduType))
        stream.write(Uint16LE.pack(header.source))

    def parseShareDataHeader(self, stream: BytesIO, controlHeader: ShareControlHeader):
        shareID = Uint32LE.unpack(stream)
        stream.read(1)
        streamID = Uint8.unpack(stream)
        uncompressedLength = Uint16LE.unpack(stream)
        pduSubtype = Uint8.unpack(stream)
        compressedType = Uint8.unpack(stream)
        compressedLength = Uint16LE.unpack(stream)
        return ShareDataHeader(controlHeader.pduType,
                               controlHeader.version, controlHeader.source,
                               shareID, streamID, uncompressedLength,
                               SlowPathDataType(pduSubtype), compressedType, compressedLength)

    def writeShareDataHeader(self, stream: BytesIO, header, dataLength):
        substream = BytesIO()
        substream.write(Uint32LE.pack(header.shareID))
        substream.write(b"\x00")
        substream.write(Uint8.pack(header.streamID))
        substream.write(Uint16LE.pack(header.uncompressedLength))
        substream.write(Uint8.pack(header.subtype))
        substream.write(Uint8.pack(header.compressedType))
        substream.write(Uint16LE.pack(header.compressedLength))
        substream = substream.getvalue()

        self.writeShareControlHeader(stream, header, dataLength + len(substream))
        stream.write(substream)

    def parseDemandActive(self, stream: BytesIO, header):
        shareID = Uint32LE.unpack(stream)
        lengthSourceDescriptor = Uint16LE.unpack(stream)
        lengthCombinedCapabilities = Uint16LE.unpack(stream)
        sourceDescriptor = stream.read(lengthSourceDescriptor)
        numberCapabilities = Uint16LE.unpack(stream)
        stream.read(2)  # Padding
        capabilitySets = stream.read(lengthCombinedCapabilities - 4)
        sessionID = Uint32LE.unpack(stream)
        parsedCapabilitySets = self.parseCapabilitySets(capabilitySets, numberCapabilities)

        return DemandActivePDU(header, shareID, sourceDescriptor, numberCapabilities,
                               capabilitySets, sessionID, parsedCapabilitySets)

    def writeDemandActive(self, stream: BytesIO, pdu: DemandActivePDU):
        Uint32LE.pack(pdu.shareID, stream)
        Uint16LE.pack(len(pdu.sourceDescriptor), stream)

        substream = BytesIO()
        self.writeCapabilitySets(pdu.parsedCapabilitySets.values(), substream)

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)  # lengthCombinedCapabilities

        stream.write(pdu.sourceDescriptor)  # sourceDescriptor
        Uint16LE.pack(len(pdu.parsedCapabilitySets.keys()), stream)  # numberCapabilities
        stream.write(b"\x00" * 2)  # pad2Octets
        stream.write(substream.getvalue())  # capabilitySets
        Uint32LE.pack(pdu.sessionID, stream)

    def parseConfirmActive(self, stream: BytesIO, header):
        shareID = Uint32LE.unpack(stream)
        originatorID = Uint16LE.unpack(stream)
        lengthSourceDescriptor = Uint16LE.unpack(stream)
        lengthCombinedCapabilities = Uint16LE.unpack(stream)
        sourceDescriptor = stream.read(lengthSourceDescriptor)
        numberCapabilities = Uint16LE.unpack(stream)
        stream.read(2)
        capabilitySetsRaw = stream.read(lengthCombinedCapabilities - 4)
        capabilitySets = self.parseCapabilitySets(capabilitySetsRaw, numberCapabilities)

        return ConfirmActivePDU(header, shareID, originatorID, sourceDescriptor,
                                numberCapabilities, capabilitySets, capabilitySetsRaw)

    def parseCapabilitySets(self, capabilitySetsRaw, numberCapabilities):
        stream = BytesIO(capabilitySetsRaw)
        capabilitySets = {}
        # Do minimum parsing for every capability
        for i in range(numberCapabilities):
            capabilitySetType = Uint16LE.unpack(stream.read(2))
            lengthCapability = Uint16LE.unpack(stream.read(2))
            capabilityData = stream.read(lengthCapability - 4)
            capability = Capability(capabilitySetType, capabilityData)
            if capabilitySetType == 6:
                # Workaround for XRDP returning bad capability type
                # ref https://github.com/GoSecure/pyrdp/issues/294
                continue
            capabilitySets[CapabilityType(capabilitySetType)] = capability

        # Fully parse the General capability set
        capabilitySets[CapabilityType.CAPSTYPE_GENERAL] = \
            self.parseGeneralCapability(capabilitySets[CapabilityType.CAPSTYPE_GENERAL].rawData)

        # Fully parse the Glyph cache capability set
        if CapabilityType.CAPSTYPE_GLYPHCACHE in capabilitySets:
            capabilitySets[CapabilityType.CAPSTYPE_GLYPHCACHE] = \
                self.parseGlyphCacheCapability(capabilitySets[CapabilityType.CAPSTYPE_GLYPHCACHE].rawData)

        # If present, fully parse the offscreen cache capability set
        if CapabilityType.CAPSTYPE_OFFSCREENCACHE in capabilitySets:
            capabilitySets[CapabilityType.CAPSTYPE_OFFSCREENCACHE] = \
                self.parseOffscreenCacheCapability(capabilitySets[CapabilityType.CAPSTYPE_OFFSCREENCACHE].rawData)

        # If present, fully parse the surface commands cache capability set
        if CapabilityType.CAPSETTYPE_SURFACE_COMMANDS in capabilitySets:
            capabilitySets[CapabilityType.CAPSETTYPE_SURFACE_COMMANDS] = \
                self.parseSurfaceCommandsCapability(capabilitySets[CapabilityType.CAPSETTYPE_SURFACE_COMMANDS].rawData)

        # Fully parse the Bitmap capability set
        capabilitySets[CapabilityType.CAPSTYPE_BITMAP] = \
            self.parseBitmapCapability(capabilitySets[CapabilityType.CAPSTYPE_BITMAP].rawData)

        # Fully parse the Order capability set
        capabilitySets[CapabilityType.CAPSTYPE_ORDER] = self.parseOrderCapability(
            capabilitySets[CapabilityType.CAPSTYPE_ORDER].rawData)

        # Fully parse the VirtualChannel capability set
        if CapabilityType.CAPSTYPE_VIRTUALCHANNEL in capabilitySets:
            capabilitySets[CapabilityType.CAPSTYPE_VIRTUALCHANNEL] = self.parseVirtualChannelCapability(
                capabilitySets[CapabilityType.CAPSTYPE_VIRTUALCHANNEL].rawData)

        # Fully parse the Pointer capability set
        if CapabilityType.CAPSTYPE_POINTER in capabilitySets:
            capabilitySets[CapabilityType.CAPSTYPE_POINTER] = self.parsePointerCapability(
                capabilitySets[CapabilityType.CAPSTYPE_POINTER].rawData)

        return capabilitySets

    def parseGeneralCapability(self, data) -> GeneralCapability:
        """
        https://msdn.microsoft.com/en-us/library/cc240549.aspx
        :param data: Raw data starting after lengthCapability
        """
        stream = BytesIO(data)
        osMajorType = Uint16LE.unpack(stream.read(2))
        osMinorType = Uint16LE.unpack(stream.read(2))
        protocolVersion = Uint16LE.unpack(stream.read(2))
        stream.read(2)  # pad2octetsA
        generalCompressionTypes = Uint16LE.unpack(stream.read(2))
        extraFlags = Uint16LE.unpack(stream.read(2))
        updateCapabilityFlag = Uint8.unpack(stream.read(1))
        remoteUnshareFlag = Uint8.unpack(stream.read(1))
        generalCompressionLevel = Uint16LE.unpack(stream.read(2))
        refreshRectSupport = Uint8.unpack(stream.read(1))
        suppressOutputSupport = Uint8.unpack(stream.read(1))

        capability = GeneralCapability(osMajorType, osMinorType, protocolVersion, generalCompressionTypes, extraFlags,
                                       updateCapabilityFlag, remoteUnshareFlag, generalCompressionLevel,
                                       refreshRectSupport, suppressOutputSupport)
        capability.rawData = data
        return capability

    def parseSurfaceCommandsCapability(self, data: bytes) -> SurfaceCommandsCapability:
        """
        https://msdn.microsoft.com/en-us/library/dd871563.aspx
        :param data: Raw data starting after lengthCapability
        """
        stream = BytesIO(data)
        cmdFlags = Uint32LE.unpack(stream)
        reserved = Uint32LE.unpack(stream)
        capability = SurfaceCommandsCapability(cmdFlags, reserved)
        capability.rawData = data
        return capability

    def writeSurfaceCommandsCapability(self, capability: SurfaceCommandsCapability, stream: BytesIO):
        substream = BytesIO()
        Uint16LE.pack(capability.capabilityType, stream)

        Uint32LE.pack(capability.cmdFlags, substream)
        Uint32LE.pack(capability.reserved, substream)

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)
        stream.write(substream.getvalue())

    def parseOffscreenCacheCapability(self, data) -> OffscreenBitmapCacheCapability:
        """
        https://msdn.microsoft.com/en-us/library/cc240550.aspx
        :param data: Raw data starting after lengthCapability
        """
        stream = BytesIO(data)
        offscreenSupportLevel = Uint32LE.unpack(stream.read(4))
        offscreenCacheSize = Uint16LE.unpack(stream.read(2))
        offscreenCacheEntries = Uint16LE.unpack(stream.read(2))

        capability = OffscreenBitmapCacheCapability(offscreenSupportLevel, offscreenCacheSize, offscreenCacheEntries)
        capability.rawData = data
        return capability

    def parseGlyphCacheCapability(self, data) -> GlyphCacheCapability:
        """
        https://msdn.microsoft.com/en-us/library/cc240565.aspx
        :param data: Raw data starting after lengthCapability
        """
        stream = BytesIO(data)
        glyphCache = stream.read(40)
        fragCache = Uint32LE.unpack(stream.read(4))
        glyphSupportLevel = Uint16LE.unpack(stream.read(2))
        stream.read(2)  # pad2octets

        capability = GlyphCacheCapability(glyphCache, fragCache, glyphSupportLevel)
        capability.rawData = data
        return capability

    def parseBitmapCapability(self, data: bytes) -> BitmapCapability:
        """
        https://msdn.microsoft.com/en-us/library/cc240554.aspx
        :param data: Raw data starting after lengthCapability
        """
        stream = BytesIO(data)
        preferredBitsPerPixel = Uint16LE.unpack(stream.read(2))
        receive1bitPerPixel = Uint16LE.unpack(stream.read(2))
        receive4bitPerPixel = Uint16LE.unpack(stream.read(2))
        receive8bitPerPixel = Uint16LE.unpack(stream.read(2))
        desktopWidth = Uint16LE.unpack(stream.read(2))
        desktopHeight = Uint16LE.unpack(stream.read(2))
        stream.read(2)  # pad2octets
        desktopResizeFlag = Uint16LE.unpack(stream.read(2))
        bitmapCompressionFlag = Uint16LE.unpack(stream.read(2))
        highColorFlags = Uint8.unpack(stream.read(1))
        drawingFlags = Uint8.unpack(stream.read(1))
        multipleRectangleSupport = Uint16LE.unpack(stream.read(2))
        # ignoring pad2octetsB

        capability = BitmapCapability(preferredBitsPerPixel, receive1bitPerPixel, receive4bitPerPixel,
                                      receive8bitPerPixel, desktopWidth, desktopHeight, desktopResizeFlag,
                                      bitmapCompressionFlag, highColorFlags, drawingFlags, multipleRectangleSupport)
        capability.rawData = data
        return capability

    def parseOrderCapability(self, data) -> OrderCapability:
        """
        https://msdn.microsoft.com/en-us/library/cc240556.aspx
        :param data: Raw data starting after lengthCapability
        """
        stream = BytesIO(data)
        terminalDescriptor = stream.read(16)
        stream.read(4)  # pad4octetsA
        desktopSaveXGranularity = Uint16LE.unpack(stream.read(2))
        desktopSaveYGranularity = Uint16LE.unpack(stream.read(2))
        stream.read(2)  # pad2octetsA
        maximumOrderLevel = Uint16LE.unpack(stream.read(2))
        numberFonts = Uint16LE.unpack(stream.read(2))
        orderFlags = Uint16LE.unpack(stream.read(2))
        orderSupport = stream.read(32)
        textFlags = Uint16LE.unpack(stream.read(2))
        orderSupportExFlags = Uint16LE.unpack(stream.read(2))
        stream.read(4)  # pad4octetsB
        desktopSaveSize = Uint32LE.unpack(stream.read(4))
        stream.read(4)  # pad2octetsC, pad2octetsD
        textANSICodePage = Uint16LE.unpack(stream.read(2))
        # ignoring pad2octetsE

        capability = OrderCapability(terminalDescriptor, desktopSaveXGranularity, desktopSaveYGranularity,
                                     maximumOrderLevel, numberFonts, orderFlags, orderSupport, textFlags,
                                     orderSupportExFlags, desktopSaveSize, textANSICodePage)
        capability.rawData = data
        return capability

    def writeConfirmActive(self, stream: BytesIO, pdu: ConfirmActivePDU):
        Uint32LE.pack(pdu.shareID, stream)
        Uint16LE.pack(pdu.originatorID, stream)
        Uint16LE.pack(len(pdu.sourceDescriptor), stream)

        substream = BytesIO()
        self.writeCapabilitySets(pdu.parsedCapabilitySets.values(), substream)

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)
        stream.write(pdu.sourceDescriptor)
        Uint16LE.pack(len(pdu.parsedCapabilitySets), stream)
        stream.write(b"\x00" * 2)  # pad2octets
        stream.write(substream.getvalue())

    def writeCapabilitySets(self, capabilitySets: List[Capability], substream):
        for capability in capabilitySets:
            # Since the general capability is fully parsed, write it back.....
            if isinstance(capability, GeneralCapability):
                self.writeGeneralCapability(capability, substream)
            elif isinstance(capability, OrderCapability):
                self.writeOrderCapability(capability, substream)
            elif isinstance(capability, BitmapCapability):
                self.writeBitmapCapability(capability, substream)
            elif isinstance(capability, SurfaceCommandsCapability):
                self.writeSurfaceCommandsCapability(capability, substream)
            elif isinstance(capability, OffscreenBitmapCacheCapability):
                self.writeOffscreenCacheCapability(capability, substream)
            elif isinstance(capability, VirtualChannelCapability):
                self.writeVirtualChannelCapability(capability, substream)
            elif isinstance(capability, PointerCapability):
                self.writePointerCapability(capability, substream)
            elif capability.capabilityType == CapabilityType.CAPSETTYPE_MULTIFRAGMENTUPDATE \
                    and isinstance(capability, MultifragmentUpdateCapability):
                self.writeMultiFragmentUpdateCapability(capability, substream)
            # Every other capability is parsed minimally.
            else:
                Uint16LE.pack(capability.capabilityType, substream)
                Uint16LE.pack(len(capability.rawData) + 4, substream)
                substream.write(capability.rawData)

    def parseError(self, stream: BytesIO, header):
        errorInfo = Uint32LE.unpack(stream)
        return SetErrorInfoPDU(header, ErrorInfo(errorInfo))

    def writeError(self, stream: BytesIO, pdu):
        Uint32LE.pack(pdu.errorInfo, stream)

    def parseSynchronize(self, stream: BytesIO, header):
        messageType = Uint16LE.unpack(stream)
        targetUser = Uint16LE.unpack(stream)
        return SynchronizePDU(header, messageType, targetUser)

    def writeSynchronize(self, stream: BytesIO, pdu):
        Uint16LE.pack(pdu.messageType, stream)
        Uint16LE.pack(pdu.targetUser, stream)

    def parseControl(self, stream: BytesIO, header):
        action = Uint16LE.unpack(stream)
        grantID = Uint16LE.unpack(stream)
        controlID = Uint32LE.unpack(stream)
        return ControlPDU(header, action, grantID, controlID)

    def writeControl(self, stream: BytesIO, pdu):
        Uint16LE.pack(pdu.action, stream)
        Uint16LE.pack(pdu.grantID, stream)
        Uint32LE.pack(pdu.grantID, stream)

    def parseInput(self, stream: BytesIO, header):
        numEvents = Uint16LE.unpack(stream)
        stream.read(2)

        parser = SlowPathInputParser()
        inputEvents = [parser.parse(stream) for _ in range(numEvents)]

        return InputPDU(header, inputEvents)

    def writeInput(self, stream: BytesIO, pdu):
        Uint16LE.pack(len(pdu.events), stream)
        stream.write(b"\x00" * 2)

        parser = SlowPathInputParser()
        for event in pdu.events:
            stream.write(parser.write(event))

    def parsePointer(self, stream: BytesIO, header):
        parser = PointerEventParser()
        event = parser.parse(stream)
        return PointerPDU(header, event)

    def writePointer(self, stream: BytesIO, pdu):
        parser = PointerEventParser()
        stream.write(parser.write(pdu.event))

    def parsePlaySound(self, stream: BytesIO, header):
        duration = Uint32LE.unpack(stream)
        frequency = Uint32LE.unpack(stream)
        return PlaySoundPDU(header, duration, frequency)

    def writePlaySound(self, stream: BytesIO, pdu):
        Uint32LE.pack(pdu.duration, stream)
        Uint32LE.pack(pdu.frequency, stream)

    def parseSuppressOutput(self, stream: BytesIO, header):
        allowDisplayUpdates = Uint8.unpack(stream)
        stream.read(3)

        if allowDisplayUpdates == 1:
            left = Uint16LE.unpack(stream)
            top = Uint16LE.unpack(stream)
            right = Uint16LE.unpack(stream)
            bottom = Uint16LE.unpack(stream)
            return SuppressOutputPDU(header, allowDisplayUpdates, left, top, right, bottom)
        else:
            return SuppressOutputPDU(header, allowDisplayUpdates, None, None, None, None)

    def writeSuppressOutput(self, stream: BytesIO, pdu):
        Uint8.pack(int(pdu.allowDisplayUpdates), stream)
        stream.write(b"\x00" * 3)

        if pdu.allowDisplayUpdates:
            Uint16LE.pack(pdu.left, stream)
            Uint16LE.pack(pdu.top, stream)
            Uint16LE.pack(pdu.right, stream)
            Uint16LE.pack(pdu.bottom, stream)

    def parseUpdate(self, stream: BytesIO, header):
        updateType = Uint16LE.unpack(stream)
        updateData = stream.read(header.uncompressedLength - 18)
        return UpdatePDU(header, updateType, updateData)

    def writeUpdate(self, stream: BytesIO, pdu: UpdatePDU):
        Uint16LE.pack(pdu.updateType, stream)
        stream.write(pdu.updateData)

    def writeGeneralCapability(self, capability: GeneralCapability, stream: BytesIO):
        """
        https://msdn.microsoft.com/en-us/library/cc240549.aspx
        """
        substream = BytesIO()
        Uint16LE.pack(capability.capabilityType, stream)
        Uint16LE.pack(capability.majorType, substream)
        Uint16LE.pack(capability.minorType, substream)
        Uint16LE.pack(capability.protocolVersion, substream)
        substream.write(b"\00" * 2)  # pad2octetsA
        Uint16LE.pack(capability.generalCompressionTypes, substream)
        Uint16LE.pack(capability.extraFlags, substream)
        Uint16LE.pack(capability.updateCapabilityFlag, substream)
        Uint16LE.pack(capability.remoteUnshareFlag, substream)
        Uint16LE.pack(capability.generalCompressionLevel, substream)
        Uint8.pack(capability.refreshRectSupport, substream)
        Uint8.pack(capability.suppressOutputSupport, substream)

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)
        stream.write(substream.getvalue())

    def writeOrderCapability(self, capability: OrderCapability, stream: BytesIO):
        substream = BytesIO()
        Uint16LE.pack(capability.capabilityType, stream)
        substream.write(capability.terminalDescriptor)
        substream.write(b"\x00"*4)
        Uint16LE.pack(capability.desktopSaveXGranularity, substream)
        Uint16LE.pack(capability.desktopSaveYGranularity, substream)
        substream.write(b"\x00" * 2)
        Uint16LE.pack(capability.maximumOrderLevel, substream)
        Uint16LE.pack(capability.numberFonts, substream)
        Uint16LE.pack(capability.orderFlags, substream)
        substream.write(capability.orderSupport)
        Uint16LE.pack(capability.textFlags, substream)
        Uint16LE.pack(capability.orderSupportExFlags, substream)
        substream.write(b"\x00" * 4)
        Uint32LE.pack(capability.desktopSaveSize, substream)
        substream.write(b"\x00" * 4)
        Uint16LE.pack(capability.textANSICodePage, substream)
        substream.write(b"\x00" * 2)

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)
        stream.write(substream.getvalue())

    def writeBitmapCapability(self, capability: BitmapCapability, stream: BytesIO):
        substream = BytesIO()
        Uint16LE.pack(capability.capabilityType, stream)

        Uint16LE.pack(capability.preferredBitsPerPixel, substream)
        Uint16LE.pack(capability.receive1BitPerPixel, substream)
        Uint16LE.pack(capability.receive4BitsPerPixel, substream)
        Uint16LE.pack(capability.receive8BitsPerPixel, substream)
        Uint16LE.pack(capability.desktopWidth, substream)
        Uint16LE.pack(capability.desktopHeight, substream)
        substream.write(b"\x00"*2)  # pad2octets
        Uint16LE.pack(capability.desktopResizeFlag, substream)
        Uint16LE.pack(capability.bitmapCompressionFlag, substream)
        Uint8.pack(capability.highColorFlags, substream)
        Uint8.pack(capability.drawingFlags, substream)
        Uint16LE.pack(capability.multipleRectangleSupport, substream)

        substream.write(b"\x00" * 2)  # pad2octetsB

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)
        stream.write(substream.getvalue())

    def writeOffscreenCacheCapability(self, capability: OffscreenBitmapCacheCapability, stream: BytesIO):
        substream = BytesIO()
        Uint16LE.pack(capability.capabilityType, stream)

        Uint32LE.pack(capability.offscreenSupportLevel, substream)
        Uint16LE.pack(capability.offscreenCacheSize, substream)
        Uint16LE.pack(capability.offscreenCacheEntries, substream)

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)
        stream.write(substream.getvalue())

    def writeMultiFragmentUpdateCapability(self, capability: MultifragmentUpdateCapability, stream: BytesIO):
        substream = BytesIO()
        Uint16LE.pack(capability.capabilityType, stream)

        Uint32LE.pack(capability.maxRequestSize, substream)

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)
        stream.write(substream.getvalue())

    def parseVirtualChannelCapability(self, data: bytes) -> VirtualChannelCapability:
        """
        https://msdn.microsoft.com/en-us/library/cc240551.aspx
        :param data: Raw data starting after lengthCapability
        """
        stream = BytesIO(data)
        flags = Uint32LE.unpack(stream)
        VCChunkSize = Uint32LE.unpack(stream) if stream.tell() != len(stream.getvalue()) else None

        return VirtualChannelCapability(flags, VCChunkSize)

    def writeVirtualChannelCapability(self, capability: VirtualChannelCapability, stream: BytesIO):
        substream = BytesIO()
        Uint16LE.pack(capability.capabilityType, stream)

        Uint32LE.pack(capability.flags, substream)

        if capability.vcChunkSize is not None:
            Uint32LE.pack(capability.vcChunkSize, substream)

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)
        stream.write(substream.getvalue())

    def parsePointerCapability(self, data: bytes) -> PointerCapability:
        """
        https://msdn.microsoft.com/en-us/library/cc240562.aspx
        :param data: Raw data starting after lengthCapability
        """
        stream = BytesIO(data)
        colorPointerFlag = Uint16LE.unpack(stream)
        colorPointerCacheSize = Uint16LE.unpack(stream)

        if len(data) == 6:
            pointerCacheSize = Uint16LE.unpack(stream)
        else:
            pointerCacheSize = None

        return PointerCapability(colorPointerFlag, colorPointerCacheSize, pointerCacheSize)

    def writePointerCapability(self, capability: PointerCapability, stream: BytesIO):
        substream = BytesIO()
        Uint16LE.pack(capability.capabilityType, stream)

        Uint16LE.pack(capability.colorPointerFlag, substream)
        Uint16LE.pack(capability.colorPointerCacheSize, substream)

        if capability.pointerCacheSize is not None:
            Uint16LE.pack(capability.pointerCacheSize, substream)

        Uint16LE.pack(len(substream.getvalue()) + 4, stream)
        stream.write(substream.getvalue())

    def parsePersistentCacheKeys(self, stream: BytesIO, header):
        num0 = Uint16LE.unpack(stream)
        num1 = Uint16LE.unpack(stream)
        num2 = Uint16LE.unpack(stream)
        num3 = Uint16LE.unpack(stream)
        num4 = Uint16LE.unpack(stream)

        total0 = Uint16LE.unpack(stream)
        total1 = Uint16LE.unpack(stream)
        total2 = Uint16LE.unpack(stream)
        total3 = Uint16LE.unpack(stream)
        total4 = Uint16LE.unpack(stream)
        bBitMask = Uint8.unpack(stream)

        stream.read(3)  # Padding

        keys = stream.read(64 * (num0 + num1 + num2 + num3 + num4))
        return PersistentCacheKeysPDU(header, num0, num1, num2, num3, num4,
                                      total0, total1, total2, total3, total4, keys, bBitMask)

    def writePersistentCacheKeys(self, s: BytesIO, pdu: PersistentCacheKeysPDU):
        # Only send the first PDU with an empty list and drop the rest.
        # TODO: Find a way to cleanly drop the entire packet instead.
        Uint16LE.pack(0, s)
        Uint16LE.pack(0, s)
        Uint16LE.pack(0, s)
        Uint16LE.pack(0, s)
        Uint16LE.pack(0, s)

        Uint16LE.pack(0, s)
        Uint16LE.pack(0, s)
        Uint16LE.pack(0, s)
        Uint16LE.pack(0, s)
        Uint16LE.pack(0, s)
        Uint8.pack(pdu.mask, s)
        s.write(b'\x00'*3)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.layer.layer import Layer
from pyrdp.parser import DeviceRedirectionParser


class DeviceRedirectionLayer(Layer):
    """
    Layer to receive and send DeviceRedirection channel (rdpdr) packets.
    """

    def __init__(self, parser = DeviceRedirectionParser()):
        super().__init__(parser)
